import json
import os

from .dcp_auth_client import DCPAuthClient


class S2STokenClient:
    def __init__(self, dcp_auth_client: 'DCPAuthClient' = None):
        self.dcp_auth_client = dcp_auth_client

    def setup_from_env_var(self, env_var_name):
        key_dict = json.loads(os.environ.get(env_var_name))
        self.dcp_auth_client = DCPAuthClient(key_dict["project_id"], key_dict)

    def setup_from_file(self, file_path):
        with open(file_path) as fh:
            service_credentials = json.load(fh)
        self.dcp_auth_client = DCPAuthClient(service_credentials["project_id"], service_credentials)

    def retrieve_token(self) -> str:
        return self.dcp_auth_client.token



from collections import namedtuple


class MigrationParser():
    """
    MigrationParser reads a json formatted list of migrations from a property_migrations file (parsed by the
    SchemaTemplate class) and captures the information in a dictionary structure.
    """

    def __init__(self, json_migrations):
        self.migrations = json_migrations

    @property
    def migrations(self):
        return self._migrations

    @migrations.setter
    def migrations(self, json_migrations):
        """
        Iterate through each of the dictionaries in json_migrations which represent a single change for a single
        property that occurred between two versions of a schema and capture the information in a dictionary structure.

        :param json_migrations: A list of dictionaries where each dictionary is a single change made to a single
                                property in a single schema version bump.
        """
        migrations = {}
        for property_migration in json_migrations:
            source_migrated_property = property_migration["source_schema"] + "." + property_migration["property"]

            if source_migrated_property in migrations:
                raise Exception(
                    f"ERROR: Property {source_migrated_property} already has migration information. Duplicates not "
                    f"expected!")

            migrations[source_migrated_property] = self._get_migration_info_of_migrated_property(property_migration)
        self._migrations = migrations

    @staticmethod
    def get_dictionary_representation(migrations):
        """
        Returns a nested dictionary representation of the given migration structure (that contains a nested
        MigrationInfo object).
        """

        return {migrated_property_name: migrated_property_info._asdict() for
                migrated_property_name, migrated_property_info in migrations.items()}

    @staticmethod
    def _get_migration_info_of_migrated_property(property_migration):
        """
        Derive the necessary migration information from the property migration JSON object (formatted as a dictionary)
        and return it as a namedtuple MigrationInfo.
        """

        if "target_schema" in property_migration and "replaced_by" in property_migration:
            _replaced_by = property_migration["target_schema"] + "." + property_migration[
                "replaced_by"]
        else:
            _replaced_by = None
        if "effective_from" in property_migration:
            _version = property_migration["effective_from"]
            _target_version = None
        elif "effective_from_source" in property_migration:
            _version = property_migration["effective_from_source"]
            _target_version = property_migration["effective_from_target"]

        MigrationInfo = namedtuple("MigrationInfo", "replaced_by version target_version")
        return MigrationInfo(replaced_by=_replaced_by, version=_version, target_version=_target_version)



#!/usr/bin/env python
"""
Description goes here
"""
__author__ = "jupp"
__license__ = "Apache 2.0"
__date__ = "04/05/2018"

from yaml import load as yaml_load

from ingest.utils import doctict
from ingest.utils.doctict import DotDict


class TabConfig:
    def __init__(self, init=None):
        self._dic = DotDict()
        self._key_label = {}
        if init:
            self._dic = DotDict(init)
            self._index()

    def load(self, input):
        stream = open(input, 'r').read()
        yaml = yaml_load(stream)
        self._dic = DotDict(yaml)
        self._index()
        return self

    def _index(self):
        if "tabs" in self._dic:
            for tab_name in self._dic["tabs"]:
                for key, value in tab_name.items():
                    dn = value["display_name"]
                    self._key_label[dn.lower()] = key
                    self._key_label[key] = key
        else:
            print("warning")
        return self

    def lookup(self, key):
        return doctict.get(self._dic, key)

    def get_key_for_label(self, label):
        return self._key_label[label.lower()]



from openpyxl import Workbook, load_workbook

from ingest.importer.spreadsheet.ingest_worksheet import IngestWorksheet
from ingest.importer.submission import Submission

SCHEMAS_WORKSHEET = 'Schemas'
SPECIAL_TABS = [SCHEMAS_WORKSHEET]


class IngestWorkbook:

    def __init__(self, workbook: Workbook):
        self.workbook = workbook

    @classmethod
    def from_file(cls, file_path, read_only=True) -> 'IngestWorkbook':
        workbook = load_workbook(filename=file_path, read_only=read_only)
        return cls(workbook)

    def get_worksheet(self, worksheet_title):
        if worksheet_title in self.workbook.sheetnames:
            return self.workbook[worksheet_title]

        if worksheet_title.lower() in self.workbook.sheetnames:
            return self.workbook[worksheet_title.lower()]

        return None

    def get_schemas(self):
        schemas = []

        worksheet = None

        if SCHEMAS_WORKSHEET in self.workbook.sheetnames:
            worksheet = self.workbook[SCHEMAS_WORKSHEET]

        if SCHEMAS_WORKSHEET.lower() in self.workbook.sheetnames:
            worksheet = self.workbook[SCHEMAS_WORKSHEET.lower()]

        if not worksheet:
            return schemas

        for row in worksheet.iter_rows(min_row=2, max_row=worksheet.max_row):
            schema_cell = row[0]
            schemas.append(schema_cell.value)
        return schemas

    def importable_worksheets(self):
        return [IngestWorksheet(worksheet) for worksheet in self.workbook.worksheets
                if worksheet.title not in SPECIAL_TABS]

    def add_entity_uuids(self, submission: Submission):
        worksheets = {}
        col_idx = 1
        entities = [entity for entity in submission.get_entities() if entity.spreadsheet_location]
        for entity in entities:
            worksheet_title = entity.spreadsheet_location.get('worksheet_title')
            row_index = entity.spreadsheet_location.get('row_index')

            if not worksheets.get(worksheet_title):
                worksheet = self.workbook[worksheet_title]
                ingest_worksheet = IngestWorksheet(worksheet=worksheet)
                worksheets[worksheet_title] = ingest_worksheet

            column_header = f'{entity.concrete_type}.uuid'

            if column_header not in ingest_worksheet.get_column_headers():
                ingest_worksheet.insert_column_with_header(column_header, col_idx)

            ingest_worksheet = worksheets.get(worksheet_title)
            ingest_worksheet.cell(row=row_index, column=col_idx).value = entity.uuid

    def add_schemas_worksheet(self, schemas):
        if SCHEMAS_WORKSHEET not in self.workbook.sheetnames:
            schema_sheet = self.workbook.create_sheet(SCHEMAS_WORKSHEET)
            schema_sheet.insert_cols(1)
            row_idx = 1
            col_idx = 1
            schema_sheet.cell(row=row_idx, column=col_idx).value = SCHEMAS_WORKSHEET
            for schema in schemas:
                row_idx += 1
                schema_sheet.cell(row=row_idx, column=col_idx).value = schema

    def save(self, file_path):
        self.workbook.save(file_path)



import logging
from typing import Optional
from os import environ
from time import sleep

import requests
from requests import HTTPError

from ingest.api.ingestapi import IngestApi
from ingest.api.stagingapi import StagingApi, FileUploadFailed
from ingest.exporter.exceptions import FileDuplication
from ingest.exporter.metadata import MetadataResource

logger = logging.getLogger(__name__)


wait_time_env = environ.get('STAGING_WAIT_TIME_MILLIS', '250')
STAGING_WAIT_TIME = float(wait_time_env) / 1000
if STAGING_WAIT_TIME < 0:
    raise ValueError('Staging wait time cannot be less than 0.')

wait_attempts_env = environ.get('STAGING_WAIT_ATTEMPTS', '5')
STAGING_WAIT_ATTEMPTS = int(wait_attempts_env)
if STAGING_WAIT_ATTEMPTS < 0:
    raise ValueError('Staging wait attempts cannot be less than 0.')


class StagingInfo:

    def __init__(self, staging_area_uuid, file_name, metadata_uuid='', cloud_url=''):
        self.staging_area_uuid = staging_area_uuid
        self.file_name = file_name
        self.metadata_uuid = metadata_uuid
        self.cloud_url = cloud_url


class StagingInfoRepository:

    def __init__(self, ingest_client: IngestApi):
        self.ingest_client = ingest_client

    def save(self, staging_info: StagingInfo) -> StagingInfo:
        try:
            self.ingest_client.create_staging_job(staging_info.staging_area_uuid, staging_info.file_name, staging_info.metadata_uuid)
        except HTTPError as http_error:
            r = http_error.response
            if r.status_code == requests.codes.conflict:
                raise FileDuplication(staging_info.staging_area_uuid, staging_info.file_name)
            else:
                raise http_error
        else:
            return staging_info

    def find_one(self, staging_area_uuid, file_name) -> Optional[StagingInfo]:
        try:
            staging_job = self.ingest_client.find_staging_job(staging_area_uuid, file_name)
        except HTTPError as http_error:
            r = http_error.response
            if r.status_code == requests.codes.not_found:
                return None
            else:
                raise http_error
        else:
            staging_area_uuid = staging_job.get('stagingAreaUuid')
            file_name = staging_job.get('stagingAreaFileName')
            cloud_url = staging_job.get('stagingAreaFileUri')
            metadata_uuid = staging_job.get('metadataUuid')

            return StagingInfo(staging_area_uuid, file_name, cloud_url=cloud_url, metadata_uuid=metadata_uuid)

    def update(self, staging_info: StagingInfo) -> StagingInfo:
        staging_job = self.ingest_client.find_staging_job(staging_info.staging_area_uuid, staging_info.file_name)
        complete_job_url = staging_job['_links']['completeStagingJob']['href']
        self.ingest_client.complete_staging_job(complete_job_url, staging_info.cloud_url)
        return staging_info

    def delete(self, staging_info: StagingInfo) -> StagingInfo:
        staging_job = self.ingest_client.find_staging_job(staging_info.staging_area_uuid, staging_info.file_name)
        url = staging_job['_links']['self']['href']
        self.ingest_client.delete_staging_job(url)
        return staging_info

    def delete_staging_locks(self, staging_area_uuid: str):
        self.ingest_client.delete_staging_jobs(staging_area_uuid)


class StagingService:

    def __init__(self, staging_client: StagingApi, staging_info_repository: StagingInfoRepository):
        self.staging_client = staging_client
        self.staging_info_repository = staging_info_repository

    def get_staging_info(self, staging_area_uuid, metadata_resource: MetadataResource) -> Optional[StagingInfo]:
        file_name = metadata_resource.get_staging_file_name()
        staging_info = self.staging_info_repository.find_one(staging_area_uuid, file_name)
        remaining_attempts = STAGING_WAIT_ATTEMPTS
        while remaining_attempts and staging_info and not staging_info.cloud_url:
            sleep(STAGING_WAIT_TIME)
            staging_info = self.staging_info_repository.find_one(staging_area_uuid, file_name)
            remaining_attempts -= 1
        if staging_info and not staging_info.cloud_url:
            raise PartialStagingInfo(staging_info)
        return staging_info

    def stage_metadata(self, staging_area_uuid, metadata_resource: MetadataResource) -> Optional[StagingInfo]:
        staging_file_name = metadata_resource.get_staging_file_name()
        staging_info = StagingInfo(staging_area_uuid, staging_file_name, metadata_uuid=metadata_resource.uuid)
        try:
            self.staging_info_repository.save(staging_info)
            formatted_type = f'metadata/{metadata_resource.metadata_type}'
            bundle_metadata = metadata_resource.to_bundle_metadata()
            self._do_stage_metadata(staging_info, formatted_type, bundle_metadata)
        except FileDuplication as file_duplication:
            logger.warning(str(file_duplication))
            staging_info = self.get_staging_info(staging_area_uuid, metadata_resource)
            if not staging_info:
                raise StagingFailed(staging_area_uuid, staging_file_name)
        return staging_info

    def _do_stage_metadata(self, staging_info, formatted_type, bundle_metadata):
        try:
            # stageFile is assumed to do (sensible) internal retries (if necessary)
            file_description = self.staging_client.stageFile(staging_info.staging_area_uuid, staging_info.file_name,
                                                             bundle_metadata, formatted_type)
            staging_info.cloud_url = file_description.url
            self.staging_info_repository.update(staging_info)
        except FileUploadFailed as staging_failed:
            logging.error(str(staging_failed))
            self.staging_info_repository.delete(staging_info)
            raise staging_failed

    def staging_area_exists(self, staging_area_uuid: str) -> bool:
        return self.staging_client.hasStagingArea(staging_area_uuid)

    def cleanup_staging_area(self, staging_area_uuid):
        self.staging_client.deleteStagingArea(staging_area_uuid)
        self.staging_info_repository.delete_staging_locks(staging_area_uuid)


class PartialStagingInfo(Exception):

    def __init__(self, staging_info: StagingInfo):
        super(PartialStagingInfo, self).__init__('Unable to return StagingInfo with full details.')
        self.staging_info = staging_info


class StagingFailed(Exception):

    def __init__(self, staging_area_uuid, file_name):
        super(StagingFailed, self).__init__(f'Staging of file "{file_name}" on staging area '
                                            f'{staging_area_uuid} failed.')



class RootSchemaException(Exception):
    """ This Exception will be used with respect to errors that occur when generating a template based on a root JSON
    object.
    """
    pass


class UnknownKeySchemaException(Exception):
    """ This Exception will be used when a key representing a field in a metadata schema cannot be mapped to a known
    property.
    """
    pass



import re
from copy import deepcopy


class MetadataParseException(Exception):
    pass


class MetadataProvenance:
    def __init__(self, document_id: str, submission_date: str, update_date: str,
                 schema_major_version: int, schema_minor_version: int):
        self.document_id = document_id
        self.submission_date = submission_date
        self.update_date = update_date
        self.schema_major_version = schema_major_version
        self.schema_minor_version = schema_minor_version

    def to_dict(self):
        return deepcopy(self.__dict__)


class MetadataResource:

    def __init__(self, metadata_type, metadata_json, uuid, dcp_version,
                 provenance: MetadataProvenance):
        self.metadata_json = metadata_json
        self.uuid = uuid
        self.dcp_version = dcp_version
        self.metadata_type = metadata_type
        self.provenance = provenance

    @staticmethod
    def from_dict(data: dict, require_provenance=True):
        try:
            metadata_json = data['content']
            uuid = data['uuid']['uuid']
            dcp_version = data['dcpVersion']
            metadata_type = data['type'].lower()
            provenance = MetadataResource._derive_provenance(data, require_provenance)
            return MetadataResource(metadata_type, metadata_json, uuid, dcp_version, provenance)
        except (KeyError, TypeError) as e:
            raise MetadataParseException(e)

    @staticmethod
    def _derive_provenance(data, require_provenance):
        try:
            provenance = MetadataResource.provenance_from_dict(data)
        except MetadataParseException:
            if require_provenance:
                raise
            else:
                provenance = None
        return provenance

    @staticmethod
    def provenance_from_dict(data: dict):
        try:
            uuid = data['uuid']['uuid']
            submission_date = data['submissionDate']
            update_date = data['updateDate']

            # Populate the major and minor schema versions from the URL in the describedBy field
            schema_semver = re.findall(r'\d+\.\d+\.\d+', data["content"]["describedBy"])[0]
            schema_major_version = int(schema_semver.split(".")[0])
            schema_minor_version = int(schema_semver.split(".")[1])

            return MetadataProvenance(uuid, submission_date, update_date, schema_major_version,
                                      schema_minor_version)
        except (KeyError, TypeError) as e:
            raise MetadataParseException(e)

    def get_staging_file_name(self):
        return f'{self.metadata_type}_{self.uuid}.json'

    def to_bundle_metadata(self) -> dict:
        bundle_metadata = dict()
        content = deepcopy(self.metadata_json)
        bundle_metadata.update(content)

        if self.provenance:
            provenance = {'provenance': self.provenance.to_dict()}
            bundle_metadata.update(provenance)
        return bundle_metadata


class MetadataService:

    def __init__(self, ingest_client):
        self.ingest_client = ingest_client

    def fetch_resource(self, resource_link: str) -> MetadataResource:
        raw_metadata = self.ingest_client.get_entity_by_callback_link(resource_link)
        return MetadataResource.from_dict(raw_metadata)



#!/usr/bin/env python
"""
Given a tabs template and list of schema URLs, will output a spreadsheet in Excel format.
"""

from .spreadsheet_builder import SpreadsheetBuilder

# TODO(maniarathi): Consolidate default values into a shared configuration file.
DEFAULT_INGEST_URL = "http://api.ingest.data.humancellatlas.org"
DEFAULT_SCHEMAS_ENDPOINT = "/schemas/search/latestSchemas"


class VanillaSpreadsheetBuilder(SpreadsheetBuilder):
    def __init__(self, output_file, hide_row=False):
        super(VanillaSpreadsheetBuilder, self).create_initial_spreadsheet(output_file, hide_row)

    def build(self, spreadsheet_tabs_template):
        tabs = spreadsheet_tabs_template.tabs

        for tab in tabs:
            for tab_name, detail in tab.items():

                worksheet = self.spreadsheet.add_worksheet(detail["display_name"])

                for column_index, column_name in enumerate(detail["columns"]):

                    formatted_column_name = self.get_user_friendly_column_name(spreadsheet_tabs_template,
                                                                               column_name, tab_name).upper()

                    if column_name.split(".")[-1] == "text":
                        desc = self.get_value_for_column(spreadsheet_tabs_template, column_name, "description")
                        required = bool(self.get_value_for_column(spreadsheet_tabs_template, column_name, "required"))
                        example_text = self.get_value_for_column(spreadsheet_tabs_template, column_name, "example")
                        guidelines = self.get_value_for_column(spreadsheet_tabs_template, column_name, "guidelines")
                    else:
                        if column_name + ".text" not in detail["columns"]:
                            desc = self.get_value_for_column(spreadsheet_tabs_template, column_name, "description")
                            required = bool(
                                self.get_value_for_column(spreadsheet_tabs_template, column_name, "required"))
                            example_text = self.get_value_for_column(spreadsheet_tabs_template, column_name, "example")
                            guidelines = self.get_value_for_column(spreadsheet_tabs_template, column_name, "guidelines")

                    if required:
                        formatted_column_name += " (Required)"

                    # set the user friendly name
                    worksheet.write(0, column_index, formatted_column_name, self.header_format)

                    column_width = len(formatted_column_name) if len(formatted_column_name) > 25 else 25

                    worksheet.set_column(column_index, column_index, column_width)

                    # set the description
                    worksheet.write(1, column_index, desc, self.desc_format)

                    # write example
                    worksheet.write(2, column_index,
                                    guidelines + ' For example: ' + example_text if example_text else guidelines,
                                    self.desc_format)

                    # set the key
                    worksheet.write(3, column_index, column_name, self.locked_format)

                    if column_name.split(".")[-1] == "ontology" or column_name.split(".")[-1] == "ontology_label":
                        worksheet.set_column(column_index, column_index, None, None, {'hidden': True})

                    if self.hidden_row:
                        worksheet.set_row(3, None, None, {'hidden': True})

                    if column_index == 0:
                        worksheet.set_row(0, 30)
                        worksheet.set_row(4, 30)
                        worksheet.write(4, column_index, "FILL OUT INFORMATION BELOW THIS ROW", self.header_format)
                    else:
                        worksheet.write(4, column_index, '', self.header_format)

        if self.include_schemas_tab:
            self.generate_and_add_schema_worksheet_to_spreadsheet(spreadsheet_tabs_template.get_schema_urls())



import logging

import requests

format = '[%(filename)s:%(lineno)s - %(funcName)20s() ] %(asctime)s - %(name)s - %(levelname)s - %(message)s'
logging.basicConfig(format=format)


class IngestSubmitter(object):

    def __init__(self, ingest_api):
        # TODO the IngestSubmitter should probably build its own instance of IngestApi
        self.ingest_api = ingest_api
        self.logger = logging.getLogger(__name__)
        self.PROGRESS_CTR = 50
        self.logger = logging.getLogger(__name__)

    def submit(self, entity_map, submission_url):
        submission = Submission(self.ingest_api, submission_url)
        submission.define_manifest(entity_map)

        entities = entity_map.get_entities()

        self._add_or_update_entities(entities, submission)

        if not submission.is_update():
            self._link_submission_to_project(entity_map, submission, submission_url)
            self._link_entities(entities, entity_map, submission)

        return submission

    def _link_submission_to_project(self, entity_map, submission, submission_url):
        project = entity_map.get_project()
        submission_envelope = self.ingest_api.get_submission(submission_url)
        submission_entity = Entity('submission_envelope',
                                   submission_envelope['uuid']['uuid'],
                                   None,
                                   is_reference=True,
                                   ingest_json=submission_envelope
                                   )
        submission.link_entity(project, submission_entity, 'submissionEnvelopes')

    def _link_entities(self, entities, entity_map, submission):
        progress = 0
        for entity in entities:
            for link in entity.direct_links:
                to_entity = entity_map.get_entity(link['entity'], link['id'])
                try:
                    submission.link_entity(entity, to_entity, relationship=link['relationship'])
                    progress = progress + 1
                    if progress % self.PROGRESS_CTR == 0 or (
                            progress == int(submission.manifest.get('expectedLinks', 0))):
                        manifest_url = self.ingest_api.get_link_from_resource(submission.manifest, 'self')
                        self.ingest_api.patch(manifest_url, {'actualLinks': progress})
                        self.logger.info(f"links progress: {progress}/ {submission.manifest.get('expectedLinks')}")

                except Exception as link_error:
                    error_message = f'''The {entity.type} with id {entity.id} could not be linked to {to_entity.type} \
                    with id {to_entity.id}.'''
                    self.logger.error(error_message)
                    self.logger.error(f'{str(link_error)}')
                    raise

    def _add_or_update_entities(self, entities, submission):
        for entity in entities:
            if entity.is_reference:
                submission.update_entity(entity)
            else:
                submission.add_entity(entity)


class EntityLinker(object):

    def __init__(self, template_manager):
        self.template_manager = template_manager
        self.process_id_ctr = 0

    def process_links_from_spreadsheet(self, entity_map):
        for entity in entity_map.get_entities():
            if entity.is_reference:
                continue

            self._validate_entity_links(entity_map, entity)
            self._generate_direct_links(entity_map, entity)

        return entity_map

    def _generate_direct_links(self, entity_map, entity):
        project = entity_map.get_project()

        # TODO Revisit if we need to link all entities to the project
        # currently, all entities are indirectly link to the project via the submission envelope
        # another issue is that protocols and files don't have links to project in ingest-core

        if project and not entity.type == 'project':
            if entity.type != 'protocol' and entity.type != 'file':
                entity.direct_links.append({
                    'entity': 'project',
                    'id': project.id,
                    'relationship': 'projects'
                })

        if project and entity.concrete_type == 'supplementary_file':
            project.direct_links.append({
                'entity': 'file',
                'id': entity.id,
                'relationship': 'supplementaryFiles'
            })

        links_by_entity = entity.links_by_entity

        linked_biomaterial_ids = links_by_entity.get('biomaterial', [])
        linked_process_id = links_by_entity['process'][0] if links_by_entity.get('process') else None
        linked_protocol_ids = links_by_entity.get('protocol', [])
        linked_file_ids = links_by_entity.get('file', [])

        linking_details = entity.linking_details

        if linked_biomaterial_ids or linked_file_ids:

            linking_process = self.link_process(entity_map, linked_process_id, linking_details)
            linking_process.direct_links.append({
                'entity': 'project',
                'id': project.id,
                'relationship': 'projects'
            })
            entity_map.add_entity(linking_process)

            # link output of process
            entity.direct_links.append({
                'entity': linking_process.type,
                'id': linking_process.id,
                'relationship': 'derivedByProcesses'
            })

            # apply all protocols to the linking process
            for linked_protocol_id in linked_protocol_ids:
                linking_process.direct_links.append({
                    'entity': 'protocol',
                    'id': linked_protocol_id,
                    'relationship': 'protocols'
                })

            # biomaterial-biomaterial
            # file-biomaterial
            for linked_biomaterial_id in linked_biomaterial_ids:
                linked_biomaterial_entity = entity_map.get_entity('biomaterial', linked_biomaterial_id)
                linked_biomaterial_entity.direct_links.append({
                    'entity': linking_process.type,
                    'id': linking_process.id,
                    'relationship': 'inputToProcesses'
                })

            # file-file
            for linked_file_id in linked_file_ids:
                linked_file_entity = entity_map.get_entity('file', linked_file_id)
                linked_file_entity.direct_links.append({
                    'entity': linking_process.type,
                    'id': linking_process.id,
                    'relationship': 'inputToProcesses'
                })

    def link_process(self, entity_map, linked_process_id, linking_details):
        if not linked_process_id:
            linked_process_id = self._generate_empty_process_id()

        linking_process = self.create_or_get_process(entity_map, linked_process_id, linking_details)

        return linking_process

    def _validate_entity_links(self, entity_map, entity):
        links_by_entity = entity.links_by_entity

        for link_entity_type, link_entity_ids in links_by_entity.items():
            for link_entity_id in link_entity_ids:
                if not link_entity_type == 'process':  # it is expected that no processes are defined in any tab,
                    # these will be created later
                    if not self._is_valid_spreadsheet_link(entity.type, link_entity_type):
                        raise InvalidLinkInSpreadsheet(entity, link_entity_type, link_entity_id)
                    if not entity_map.get_entity(link_entity_type, link_entity_id):
                        raise LinkedEntityNotFound(entity, link_entity_type, link_entity_id)
                    if not entity_map.get_entity(link_entity_type, link_entity_id):
                        raise LinkedEntityNotFound(entity, link_entity_type, link_entity_id)

                if link_entity_type == 'process' and not len(link_entity_ids) == 1:
                    raise MultipleProcessesFound(entity, link_entity_ids)

    def create_or_get_process(self, entity_map, process_id, linking_details):
        process = entity_map.get_entity('process', process_id)

        if not process:
            process = self.create_process(process_id, linking_details)

        return process

    @staticmethod
    def _is_valid_spreadsheet_link(from_entity_type, to_entity_type):
        VALID_ENTITY_LINKS_MAP = [
            'biomaterial-biomaterial',
            'file-biomaterial',
            'file-file',
            'biomaterial-process',
            'biomaterial-protocol',
            'file-process',
            'file-protocol',
        ]
        link_key = from_entity_type + '-' + to_entity_type

        return link_key in VALID_ENTITY_LINKS_MAP

    def create_process(self, process_id, linking_details):
        schema_type = 'process'
        described_by = self.template_manager.get_schema_url(schema_type)

        if linking_details:
            if not linking_details.get('process_core'):
                linking_details['process_core'] = {}

            linking_details['process_core']['process_id'] = process_id
            linking_details['schema_type'] = schema_type
            linking_details['describedBy'] = described_by
        else:
            process_core = {'process_id': process_id}
            linking_details = {
                "process_core": process_core,
                "schema_type": schema_type,
                "describedBy": described_by
            }

        process = Entity(
            entity_type='process',
            entity_id=process_id,
            content=linking_details
        )

        return process

    def _generate_empty_process_id(self):
        self.process_id_ctr += 1

        return 'process_id_' + str(self.process_id_ctr)


class Entity(object):
    def __init__(self, entity_type, entity_id, content, ingest_json=None, links_by_entity=None,
                 direct_links=None, is_reference=False, linking_details=None, concrete_type=None,
                 spreadsheet_location=None):
        self.type = entity_type
        self.id = entity_id
        self.content = content
        self._prepare_links_by_entity(links_by_entity)
        self._prepare_direct_links(direct_links)
        self._prepare_linking_details(linking_details)
        self.ingest_json = ingest_json
        self.is_reference = is_reference
        self.concrete_type = concrete_type
        self.spreadsheet_location = spreadsheet_location

    def _prepare_links_by_entity(self, links_by_entity):
        self.links_by_entity = {}
        if links_by_entity is not None:
            self.links_by_entity.update(links_by_entity)

    def _prepare_direct_links(self, direct_links):
        self.direct_links = []
        if direct_links is not None:
            self.direct_links.extend(direct_links)

    def _prepare_linking_details(self, linking_details):
        self.linking_details = {}
        if linking_details is not None:
            self.linking_details.update(linking_details)

    @property
    def uuid(self):
        return self.ingest_json.get('uuid', {}).get('uuid')

    @property
    def url(self):
        return self.ingest_json['_links']['self']['href']


class Submission(object):
    ENTITY_LINK = {
        'biomaterial': 'biomaterials',
        'process': 'processes',
        'file': 'files',
        'protocol': 'protocols',
        'project': 'projects',
        'submission_envelope': 'submissionEnvelopes'
    }

    def __init__(self, ingest_api, submission_url):
        self.ingest_api = ingest_api
        self.submission_url = submission_url
        self.metadata_dict = {}
        self.manifest = None
        self.logger = logging.getLogger(__name__)

    def is_update(self):
        submission = self.ingest_api.get_submission(self.submission_url)
        return submission.get('isUpdate', False)

    def get_submission_url(self):
        return self.submission_url

    def add_entity(self, entity: Entity):
        self._create_entity(entity)
        return entity

    def update_entity(self, entity: Entity):
        # TODO updating files is not supported yet
        # Do not update if the content is None or empty
        # content is empty for linked external references
        # TODO consider adding an attribute rather than checking content
        try:
            if entity.content and entity.type != 'file':
                self._create_entity(entity, entity.id)
        except requests.HTTPError as e:
            error = f'Failed to create update entity: {e.response.text}'
            if e.response.status_code == requests.codes.bad_request:
                self.logger.warning(error)
            else:
                raise
        return entity

    def _create_entity(self, entity: Entity, uuid=None):
        link_name = self.ENTITY_LINK[entity.type]

        # TODO: how to get filename?!!!
        if entity.type == 'file':
            file_name = entity.content['file_core']['file_name']
            response = self.ingest_api.create_file(self.submission_url,
                                                   file_name,
                                                   entity.content,
                                                   uuid=uuid)
        elif entity.type == 'project':
            response = self.ingest_api.create_project(self.submission_url,
                                                      entity.content,
                                                      uuid=uuid)
        else:
            response = self.ingest_api.create_entity(self.submission_url,
                                                     entity.content,
                                                     link_name, uuid=uuid)
        entity.ingest_json = response
        self.metadata_dict[entity.type + '.' + entity.id] = entity

        return response

    def get_entity(self, entity_type, id):
        key = entity_type + '.' + id
        return self.metadata_dict[key]

    def link_entity(self, from_entity, to_entity, relationship):
        if from_entity.is_reference and not from_entity.ingest_json:
            from_entity.ingest_json = self.ingest_api.get_entity_by_uuid(self.ENTITY_LINK[from_entity.type],
                                                                         from_entity.id)

        if to_entity.is_reference and not to_entity.ingest_json:
            to_entity.ingest_json = self.ingest_api.get_entity_by_uuid(self.ENTITY_LINK[to_entity.type], to_entity.id)

        from_entity_ingest = from_entity.ingest_json
        to_entity_ingest = to_entity.ingest_json
        self.ingest_api.link_entity(from_entity_ingest, to_entity_ingest, relationship)

    def define_manifest(self, entity_map):
        # TODO provide a better way to serialize
        manifest_json = {
            'totalCount': entity_map.count_total(),
            'expectedBiomaterials': entity_map.count_entities_of_type('biomaterial'),
            'expectedProcesses': entity_map.count_entities_of_type('process'),
            'expectedFiles': entity_map.count_entities_of_type('file'),
            'expectedProtocols': entity_map.count_entities_of_type('protocol'),
            'expectedProjects': entity_map.count_entities_of_type('project'),
            'expectedLinks': entity_map.count_links(),
            'actualLinks': 0
        }

        self.manifest = self.ingest_api.create_submission_manifest(self.submission_url, manifest_json)
        return self.manifest

    def get_entities(self):
        return self.metadata_dict.values()


class EntityMap(object):

    def __init__(self, *entities):
        self.entities_dict_by_type = {}
        if entities is not None:
            for entity in entities:
                self.add_entity(entity)

    @staticmethod
    def load(entity_json):
        dictionary = EntityMap()

        for entity_type, entities_dict in entity_json.items():
            for entity_id, entity_body in entities_dict.items():

                external_links = entity_body.get('external_links_by_entity')

                if not external_links:
                    external_links = {}

                for external_link_type, external_link_uuids in external_links.items():
                    for entity_uuid in external_link_uuids:
                        external_link_entity = Entity(entity_type=external_link_type,
                                                      entity_id=entity_uuid,
                                                      content=None,
                                                      spreadsheet_location=entity_body.get('spreadsheet_location'),
                                                      is_reference=True)

                        dictionary.add_entity(external_link_entity)

                        if not entity_body.get('links_by_entity'):
                            entity_body['links_by_entity'] = {}

                        if not entity_body['links_by_entity'].get(external_link_type):
                            entity_body['links_by_entity'][external_link_type] = []

                        entity_body['links_by_entity'][external_link_type].append(entity_uuid)

                entity = Entity(entity_type=entity_type,
                                entity_id=entity_id,
                                content=entity_body.get('content'),
                                links_by_entity=entity_body.get('links_by_entity', {}),
                                is_reference=entity_body.get('is_reference', False),
                                linking_details=entity_body.get('linking_details', {}),
                                concrete_type=entity_body.get('concrete_type'),
                                spreadsheet_location=entity_body.get(
                                    'spreadsheet_location'))

                dictionary.add_entity(entity)

        return dictionary

    def get_entity_types(self):
        return list(self.entities_dict_by_type.keys())

    def get_entities_of_type(self, type):
        entities = []

        entities_dict = self.entities_dict_by_type.get(type, {})

        for entity_id, entity in entities_dict.items():
            entities.append(entity)

        return entities

    def get_new_entities_of_type(self, type):
        entities = []
        entities_dict = self.entities_dict_by_type.get(type, {})
        for entity_id, entity in entities_dict.items():
            if not entity.is_reference:
                entities.append(entity)

        return entities

    def get_entity(self, type, id):
        if self.entities_dict_by_type.get(type) and self.entities_dict_by_type[type].get(id):
            return self.entities_dict_by_type[type][id]

    def add_entity(self, entity):
        entities_of_type = self.entities_dict_by_type.get(entity.type)
        if not entities_of_type:
            self.entities_dict_by_type[entity.type] = {}
            entities_of_type = self.entities_dict_by_type.get(entity.type)
        entities_of_type[entity.id] = entity

    def get_entities(self):
        all_entities = []
        for entity_type, entities_dict in self.entities_dict_by_type.items():
            all_entities.extend(entities_dict.values())
        return all_entities

    def get_new_entities(self):
        all_entities = []
        for entity_type, entities_dict in self.entities_dict_by_type.items():
            for entity_id, entity in entities_dict.items():
                if not entity.is_reference:
                    all_entities.append(entity)
        return all_entities

    def get_project(self):
        project_id = list(self.entities_dict_by_type.get('project').keys())[0]
        return self.get_entity('project', project_id)

    def count_total(self):
        return len(self.get_entities())

    def count_entities_of_type(self, type):
        return len(self.get_new_entities_of_type(type))

    def count_links(self):
        count = 0
        for entity in self.get_entities():
            count = count + len(entity.direct_links)
        return count


class Error(Exception):
    def __init__(self, code, message):
        super(Error, self).__init__(message)
        self.code = code
        self.message = message


class InvalidEntityIngestLink(Error):
    def __init__(self, from_entity, to_entity):
        message = f'It is not possible to link a {from_entity.type} to {to_entity.type} in ingest database.'
        super(InvalidEntityIngestLink, self).__init__('InvalidEntityIngestLink', message)
        self.from_entity = from_entity
        self.to_entity = to_entity


class InvalidLinkInSpreadsheet(Error):
    def __init__(self, from_entity, link_entity_type, link_entity_id):
        message = f'It is not possible to link a {from_entity.type} to {link_entity_type} in the spreadsheet.'
        super(InvalidLinkInSpreadsheet, self).__init__('InvalidLinkInSpreadsheet', message)
        self.from_entity = from_entity
        self.link_entity_type = link_entity_type
        self.link_entity_id = link_entity_id


class LinkedEntityNotFound(Error):
    def __init__(self, from_entity, entity_type, id):
        message = f'A link from a {from_entity.type} with id {from_entity.id} to a {entity_type} with id, ' \
            f'"{id}", is not found in the spreadsheet.'
        super(LinkedEntityNotFound, self).__init__('LinkedEntityNotFound', message)
        self.entity = entity_type
        self.id = id


class MultipleProcessesFound(Error):
    def __init__(self, from_entity, process_ids):
        message = f'Multiple processes are linked {from_entity.type} in the spreadsheet: {process_ids}.'
        super(MultipleProcessesFound, self).__init__('MultipleProcessesFound', message)

        self.process_ids = process_ids
        self.from_entity = from_entity


class SubmissionError(Error):
    pass



import re

from ingest.api import utils
from ingest.api.dssapi import DssApi
from ingest.exporter.metadata import MetadataResource

_BUNDLE_FILE_TYPE_DATA = 'data'
_BUNDLE_FILE_TYPE_LINKS = 'links'

_metadata_type_attr_map = {
    'biomaterial': 'fileBiomaterialMap',
    'file': 'fileFilesMap',
    'process': 'fileProcessMap',
    'project': 'fileProjectMap',
    'protocol': 'fileProtocolMap',
}


class BundleManifest:

    def __init__(self, bundleUuid=None, envelopeUuid=None, bundleVersion=None):
        self.bundleUuid = bundleUuid
        self.envelopeUuid = envelopeUuid if envelopeUuid is not None else {}
        self.bundleVersion = bundleVersion
        self.dataFiles = []
        self.fileBiomaterialMap = {}
        self.fileProcessMap = {}
        self.fileFilesMap = {}
        self.fileProjectMap = {}
        self.fileProtocolMap = {}

    def add_bundle_file(self, metadata_type, entry: dict):
        if metadata_type == _BUNDLE_FILE_TYPE_LINKS:
            pass  # we don't want to track links.json in BundleManifests
        elif metadata_type == _BUNDLE_FILE_TYPE_DATA:
            self.dataFiles.extend(entry.keys())
        else:
            attr_mapping = _metadata_type_attr_map.get(metadata_type)
            if attr_mapping:
                file_map = getattr(self, attr_mapping)
                file_map.update(entry)
            else:
                raise KeyError(f'Cannot map unknown metadata type [{metadata_type}].')


_DSS_CONTENT_TYPE_PATTERN = re.compile('.*dcp-type="?(metadata/)?(?P<data_type>\\w+)"?.*')

_DSS_CONTENT_TYPE_TEMPLATE = 'application/json; dcp-type="metadata/{0}"'


class BundleParseFromSourceException(Exception):
    pass


class Bundle:

    def __init__(self, uuid: str, files: dict, version: str, creator_uid: int):
        self.uuid = uuid
        self.files = files
        self.version = version
        self.creator_uid = creator_uid

    @staticmethod
    def bundle_from_source(source: dict):
        try:
            bundle_source = source['bundle']

            uuid = bundle_source['uuid']
            version = bundle_source['version']
            creator_uid = bundle_source['creator_uid']
            files = Bundle._prepare_file_map(bundle_source['files'])
            return Bundle(uuid, files, version, creator_uid)
        except (KeyError, TypeError) as e:
            raise BundleParseFromSourceException(e)

    @staticmethod
    def _prepare_file_map(bundle_files: list):
        return {file['uuid']: file for file in bundle_files}

    def get_version(self):
        return self.version

    def get_file(self, uuid):
        return self.files.get(uuid)

    def get_files(self):
        return list(self.files.values())

    def count_files(self):
        return len(self.files)

    def update_version(self, version):
        self.version = version

    def update_file(self, metadata_resource: MetadataResource):
        target_file = self.get_file(metadata_resource.uuid)
        target_file['version'] = utils.to_dss_version(metadata_resource.dcp_version)
        target_file['content-type'] = _DSS_CONTENT_TYPE_TEMPLATE.format(
            metadata_resource.metadata_type)

    def generate_manifest(self, envelope_uuid) -> BundleManifest:
        manifest = BundleManifest(bundleUuid=self.uuid, envelopeUuid=envelope_uuid,
                                  bundleVersion=self.get_version())
        for file_uuid, file in self.files.items():
            pattern_match = _DSS_CONTENT_TYPE_PATTERN.match(file.get('content-type'))
            if pattern_match:
                content_type = pattern_match.group('data_type')
                mapping = {file_uuid: [file_uuid]}
                manifest.add_bundle_file(content_type, mapping)
        return manifest


class BundleService:

    def __init__(self, dss_client: DssApi):
        self.dss_client = dss_client

    def fetch(self, uuid: str, version=None) -> Bundle:
        bundle_source = self.dss_client.get_bundle(uuid, version)
        return Bundle.bundle_from_source(bundle_source)

    def update(self, bundle: Bundle, staging_details: list):
        cloud_url_map = {info.metadata_uuid: info.cloud_url for info in staging_details}
        bundle_files = bundle.get_files()
        for uuid, cloud_url in cloud_url_map.items():
            file = bundle.get_file(uuid)
            self.dss_client.put_file(None, {'url': cloud_url, 'dss_uuid': uuid,
                                            'update_date': file.get('version')})
        self.dss_client.put_bundle(bundle.uuid, bundle.get_version(), bundle_files)



import re

from xlsxwriter.worksheet import Worksheet

MODULE_TITLE_PATTERN = re.compile(r'^(?P<main_label>\w+( \w+)*)( - (?P<field_name>\w+([ -]\w+)*))?')

HEADER_ROW_IDX = 4
START_DATA_ROW = 6


class IngestWorksheet(object):

    def __init__(self, worksheet: Worksheet, header_row_idx=HEADER_ROW_IDX):
        self._worksheet = worksheet
        self._header_row_idx = header_row_idx

    @staticmethod
    def is_empty(row):
        return all(cell.value is None for cell in row)

    @property
    def title(self):
        return self._worksheet.title

    def get_column_headers(self):
        rows = self._worksheet.iter_rows(min_row=self._header_row_idx, max_row=self._header_row_idx)
        header_row = next(rows)

        headers = []
        for cell in header_row:
            if cell.value is None:
                continue

            cell_value = cell.value.strip()
            headers.append(cell_value)

        return headers

    def get_data_rows(self, start_row=START_DATA_ROW, end_row=None):
        headers = self.get_column_headers()
        max_row = end_row or self.compute_max_row()
        rows = self._worksheet.iter_rows(min_row=start_row, max_row=max_row)
        rows = [row[:len(headers)] for row in rows if not self.is_empty(row)]
        rows = [IngestRow(self._worksheet.title, START_DATA_ROW + index, row) for index, row in enumerate(rows)]
        return rows

    # NOTE: there are no tests around this because it's too complicated to
    # setup the scenario where the worksheet returns an erroneous max_row value
    def compute_max_row(self):
        max_row = self._worksheet.max_row
        if max_row is None:
            self._worksheet.calculate_dimension(force=True)
            max_row = self._worksheet.max_row
        return max_row

    def is_module_tab(self):
        match = MODULE_TITLE_PATTERN.match(self.title)
        return bool(match and match.group('field_name'))

    def get_module_field_name(self):
        match = MODULE_TITLE_PATTERN.match(self.title)
        field_name = match.group('field_name')
        if field_name:
            field_name = re.sub(r'[\s-]', '_', field_name.lower())
        return field_name

    def insert_column_with_header(self, header, col_idx):
        self._worksheet.insert_cols(col_idx)
        self._worksheet.cell(row=self._header_row_idx, column=col_idx).value = header

    def cell(self, row, column):
        return self._worksheet.cell(row=row, column=column)


class IngestRow(object):
    def __init__(self, worksheet_title, index, values):
        self.values = values or []
        self.index = index or None  # starts at 1
        self.worksheet_title = worksheet_title or None



class BundleDSSError(Exception):
    """There was a failure in bundle creation in DSS."""


class BundleFileUploadError(Exception):
    """There was a failure in bundle file upload."""


class FileDSSError(Exception):
    """There was a failure in file creation in DSS."""


class InvalidBundleError(Exception):
    """There was a failure in bundle validation."""


class MultipleProjectsError(Exception):
    """A process should only have one project linked."""


class NoUploadAreaFoundError(Exception):
    """Export couldn't be as no upload area found"""


class FileDuplication(Exception):

    def __init__(self, staging_area_uuid, file_name):
        message = f'A file with name "{file_name}" already exists in staging area ' \
                  f'{staging_area_uuid}.'
        super(FileDuplication, self).__init__(message)



#!/usr/bin/env python
import json
import logging
import os
import re
import time
import uuid
from copy import deepcopy

import polling
import requests

import ingest.exporter.bundle
from ingest.api.dssapi import DssApi, BundleAlreadyExist
from ingest.api.ingestapi import IngestApi
from ingest.exporter.metadata import MetadataResource
from ingest.exporter.staging import StagingService, StagingInfo
from ingest.utils.IngestError import ExporterError
from .exceptions import BundleDSSError, BundleFileUploadError, FileDSSError, MultipleProjectsError, \
    NoUploadAreaFoundError

DEFAULT_INGEST_URL = os.environ.get('INGEST_API', 'http://api.ingest.dev.data.humancellatlas.org')
DEFAULT_STAGING_URL = os.environ.get('STAGING_API', 'http://upload.dev.data.humancellatlas.org')
DEFAULT_DSS_URL = os.environ.get('DSS_API', 'http://dss.dev.data.humancellatlas.org')


class IngestExporter:
    def __init__(self, ingest_api: IngestApi, dss_api: DssApi, staging_service: StagingService, dry_run=False,
                 output_directory=None):
        format = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
        logging.basicConfig(format=format)
        self.logger = logging.getLogger(__name__)

        self.dry_run = dry_run
        self.output_dir = output_directory

        self.dss_api = dss_api
        self.ingest_api = ingest_api
        self.staging_service = staging_service
        self.related_entities_cache = {}

    def export_bundle(self, bundle_uuid, bundle_version, submission_uuid, process_uuid):
        start_time = time.time()
        self.related_entities_cache = {}
        saved_bundle_uuid = None

        if not self.dry_run and not self.staging_service.staging_area_exists(submission_uuid):
            error_message = "Can't do export as no upload area has been created."
            raise NoUploadAreaFoundError(error_message)

        self.logger.info("Export bundle for process with UUID %s", process_uuid)

        self.logger.info('Retrieving all process information...')

        process = self.ingest_api.get_entity_by_uuid('processes', process_uuid)
        process_info = self.get_all_process_info(process)

        self.logger.info('Generating bundle files...')
        submission = self.ingest_api.get_entity_by_uuid('submissionEnvelopes', submission_uuid)
        is_indexed = submission['triggersAnalysis']

        metadata_by_type = self.get_metadata_by_type(process_info)
        files_by_type = self.prepare_metadata_files(metadata_by_type, process_info, is_indexed)

        links = self.bundle_links(process_info.links.get_links())
        links_file_uuid = str(uuid.uuid4())
        files_by_type['links'] = list()

        original_links_doc = {
            'uuid': {
                'uuid': links_file_uuid
            },
            'type': 'links',
            'content': links,
            'dcpVersion': None
        }

        files_by_type['links'].append({
            'original_doc': original_links_doc,
            'content': links,
            'content_type': 'metadata/{0}'.format('links'),
            'indexed': is_indexed,
            'dss_filename': 'links.json',
            'dss_uuid': links_file_uuid,
            'upload_filename': 'links_' + links_file_uuid + '.json'
        })

        # restructure bundle manifest
        bundle_manifest = ingest.exporter.bundle.BundleManifest()
        bundle_manifest.envelopeUuid = submission_uuid
        bundle_manifest.bundleUuid = bundle_uuid
        bundle_manifest.bundleVersion = bundle_version
        bundle_manifest = self.create_bundle_manifest(bundle_manifest, files_by_type)

        self.logger.info('Generating bundle files...')

        if self.dry_run:
            self.logger.info('Export is using dry run mode.')
            self.logger.info('Dumping bundle files...')

            for metadata_type in ['project', 'biomaterial', 'process', 'protocol', 'file', 'links']:
                for metadata_doc in files_by_type[metadata_type]:
                    bundle_file = metadata_doc
                    filename = bundle_file['upload_filename']
                    content = bundle_file['content']
                    output_dir = self.output_dir if self.output_dir else bundle_manifest.bundleUuid
                    self.dump_to_file(json.dumps(content, indent=4), filename, output_dir=output_dir)

            self.logger.info("Dry run for bundle %s", bundle_manifest.bundleUuid)
            self.logger.info("Execution Time: %d seconds", time.time() - start_time)
        else:
            self.logger.info('Uploading metadata files...')
            try:
                self.upload_metadata_files(submission_uuid, files_by_type)
            except Exception as bundle_error:
                submission_url = self._extract_submission_url(submission)
                if submission_url:
                    self.ingest_api.create_submission_error(submission_url, ExporterError(str(bundle_error)).getJSON())
                raise

            metadata_files = self.get_metadata_files(files_by_type)
            data_files = self.get_data_files(metadata_by_type['file'])
            bundle_files = metadata_files + data_files

            bundle_manifest.dataFiles = list()
            bundle_manifest.dataFiles = [data_file['dss_uuid'] for data_file in data_files]
            self.logger.info('Saving files in DSS...')

            try:
                created_files = self.put_files_in_dss(bundle_uuid, bundle_files, process_info)

                # check all created files
                self.logger.info('Verifying if all files get successfully copied to DSS...')
                self.verify_files(created_files)

                self.logger.info('Saving bundle in DSS...')
                self.put_bundle_in_dss(bundle_uuid, bundle_version, created_files)

                self.logger.info('Saving bundle manifest...')
                self.ingest_api.create_bundle_manifest(bundle_manifest)

                saved_bundle_uuid = bundle_manifest.bundleUuid

                self.logger.info("Bundle %s was successfully created!", bundle_uuid)
                self.logger.info("Execution Time: %d seconds", time.time() - start_time)
            except BundleAlreadyExist:
                raise
            except Exception as unresolvable_exception:
                submission_url = self._extract_submission_url(submission)
                if submission_url:
                    self.ingest_api.create_submission_error(
                        submission_url,
                        ExporterError(str(unresolvable_exception)).getJSON()
                    )
                raise
        return saved_bundle_uuid

    def _extract_submission_url(self, submission_json):
        submission_url = None
        submission_links = submission_json.get('_links')
        if submission_links:
            submission_url = submission_links.get('self').get('href')
        return submission_url

    def get_metadata_by_type(self, process_info: 'ProcessInfo') -> dict:
        #  given a ProcessInfo, pull out all the metadata and return as a map of UUID->metadata documents
        simplified = dict()
        simplified['process'] = dict(process_info.derived_by_processes)
        simplified['biomaterial'] = dict(process_info.input_biomaterials)
        simplified['protocol'] = dict(process_info.protocols)
        simplified['file'] = dict(process_info.derived_files)
        simplified['file'].update(process_info.input_files)
        simplified['file'].update(process_info.supplementary_files)

        simplified['project'] = dict()
        simplified['project'][process_info.project['uuid']['uuid']] = process_info.project

        return simplified

    def get_all_process_info(self, process):
        process_info = ProcessInfo()
        process_info.input_bundle = self.get_input_bundle(process)

        process_info.project = self.get_project_info(process)

        if not process_info.project:  # get from input bundle
            project_uuid_lists = list(process_info.input_bundle['fileProjectMap'].values())

            if not project_uuid_lists and not project_uuid_lists[0]:
                raise Exception('Input bundle manifest has no list of project uuid.')  # very unlikely to happen

            project_uuid = project_uuid_lists[0][0]
            process_info.project = self.ingest_api.get_project_by_uuid(project_uuid)

        self.recurse_process(process, process_info)

        if process_info.project:
            supplementary_files = self.ingest_api.get_related_entities('supplementaryFiles', process_info.project,
                                                                       'files')
            for supplementary_file in supplementary_files:
                uuid = supplementary_file['uuid']['uuid']
                process_info.supplementary_files[uuid] = supplementary_file

        return process_info

    def get_project_info(self, process):
        projects = list(self.ingest_api.get_related_entities('projects', process, 'projects'))

        if len(projects) > 1:
            raise MultipleProjectsError('Can only be one project in bundle')

        # TODO add checking for project only on an assay process
        # TODO an analysis process may have no link to a project

        return projects[0] if projects else None

    # get all related info of a process
    def recurse_process(self, process, process_info):
        uuid = process['uuid']['uuid']
        process_info.derived_by_processes[uuid] = process

        # get all derived by processes using input biomaterial and input files
        derived_by_processes = []

        # wrapper process has the links to input biomaterials and derived files to check if a process is an assay
        input_biomaterials = self.get_related_entities('inputBiomaterials', process, 'biomaterials')
        for input_biomaterial in input_biomaterials:
            uuid = input_biomaterial['uuid']['uuid']
            process_info.input_biomaterials[uuid] = input_biomaterial
            derived_by_processes.extend(
                self.get_related_entities('derivedByProcesses', input_biomaterial, 'processes'))

        input_files = self.get_related_entities('inputFiles', process, 'files')
        for input_file in input_files:
            uuid = input_file['uuid']['uuid']
            process_info.input_files[uuid] = input_file
            derived_by_processes.extend(
                self.get_related_entities('derivedByProcesses', input_file, 'processes'))

        derived_biomaterials = self.get_related_entities('derivedBiomaterials', process, 'biomaterials')
        derived_files = self.get_related_entities('derivedFiles', process, 'files')

        process_uuid = process['uuid']['uuid']

        protocols = self.get_related_entities('protocols', process, 'protocols')
        for protocol in protocols:
            uuid = protocol['uuid']['uuid']
            process_info.protocols[uuid] = protocol

        for derived_file in derived_files:
            uuid = derived_file['uuid']['uuid']
            process_info.derived_files[uuid] = derived_file

        if input_biomaterials:
            if derived_files:
                process_info.links.add_link({
                    'process': process_uuid,
                    'inputs': [input_biomaterial['uuid']['uuid'] for input_biomaterial in input_biomaterials],
                    'input_type': 'biomaterial',
                    'outputs': [derived_file['uuid']['uuid'] for derived_file in derived_files],
                    'output_type': 'file',
                    'protocols': [
                        {
                            'protocol_type': self.get_concrete_entity_type(protocol),
                            'protocol_id': protocol['uuid']['uuid']
                        } for protocol in protocols
                    ]
                })

            if derived_biomaterials:
                process_info.links.add_link({
                    'process': process_uuid,
                    'inputs': [input_biomaterial['uuid']['uuid'] for input_biomaterial in input_biomaterials],
                    'input_type': 'biomaterial',
                    'outputs': [derived_biomaterial['uuid']['uuid'] for derived_biomaterial in derived_biomaterials],
                    'output_type': 'biomaterial',
                    'protocols': [
                        {
                            'protocol_type': self.get_concrete_entity_type(protocol),
                            'protocol_id': protocol['uuid']['uuid']
                        } for protocol in protocols
                    ]
                })

        if input_files and derived_files:
            process_info.links.add_link({
                'process': process_uuid,
                'inputs': [input_file['uuid']['uuid'] for input_file in input_files],
                'input_type': 'file',
                'outputs': [derived_file['uuid']['uuid'] for derived_file in derived_files],
                'output_type': 'file',
                'protocols': [
                    {
                        'protocol_type': self.get_concrete_entity_type(protocol),
                        'protocol_id': protocol['uuid']['uuid']
                    } for protocol in protocols
                ]
            })

        for derived_by_process in derived_by_processes:
            self.recurse_process(derived_by_process, process_info)

    def get_related_entities(self, relationship, entity, entity_type):
        entity_uuid = entity['uuid']['uuid']

        if self.related_entities_cache.get(entity_uuid) and self.related_entities_cache.get(entity_uuid).get(
                relationship):
            return self.related_entities_cache.get(entity_uuid).get(relationship)

        related_entities = list(self.ingest_api.get_related_entities(relationship, entity, entity_type))

        if not self.related_entities_cache.get(entity_uuid):
            self.related_entities_cache[entity_uuid] = {}

        if not self.related_entities_cache.get(entity_uuid).get(relationship):
            self.related_entities_cache[entity_uuid][relationship] = []

        self.related_entities_cache[entity_uuid][relationship] = related_entities

        return related_entities

    def get_input_bundle(self, process):
        bundle_manifests = list(
            self.ingest_api.get_related_entities('inputBundleManifests', process, 'bundleManifests'))

        return bundle_manifests[0] if bundle_manifests else None

    def prepare_metadata_files(self, metadata_info, process_info, is_indexed=True) -> 'dict':
        metadata_files_by_type = dict()

        for entity_type in ['biomaterial', 'file', 'project', 'protocol', 'process']:
            metadata_files_by_type[entity_type] = list()
            concrete_type_ctr = dict()
            for (metadata_uuid, doc) in metadata_info[entity_type].items():
                concrete_type = self.get_concrete_entity_type(doc)

                if concrete_type not in concrete_type_ctr:
                    concrete_type_ctr[concrete_type] = 0
                else:
                    concrete_type_ctr[concrete_type] = concrete_type_ctr[concrete_type] + 1

                file_name = '{0}_{1}.json'.format(concrete_type, concrete_type_ctr[concrete_type])
                upload_filename = '{0}_{1}.json'.format(concrete_type, metadata_uuid)

                prepared_doc = {
                    'original_doc': doc,
                    'content': self.bundle_metadata(doc, metadata_uuid),
                    'content_type': 'metadata/{0}'.format(entity_type),
                    'indexed': is_indexed,
                    'dss_filename': file_name,
                    'dss_uuid': metadata_uuid,
                    'upload_filename': upload_filename,
                    'update_date': doc['updateDate'],
                    'is_from_input_bundle': self._is_from_input_bundle(entity_type, metadata_uuid,
                                                                       process_info.input_bundle)
                }

                metadata_files_by_type[entity_type].append(prepared_doc)

        return metadata_files_by_type

    def _is_from_input_bundle(self, entity_type, metadata_uuid, input_bundle):

        field = {
            'biomaterial': 'fileBiomaterialMap',
            'process': 'fileProcessMap',
            'file': 'fileFilesMap',
            'project': 'fileProjectMap',
            'protocol': 'fileProtocolMap',
        }

        return input_bundle and input_bundle[field[entity_type]].get(metadata_uuid)

    def bundle_metadata(self, metadata_doc, uuid):
        provenance_core = dict()
        provenance_core['document_id'] = uuid
        provenance_core['submission_date'] = metadata_doc['submissionDate']
        provenance_core['update_date'] = metadata_doc['updateDate']

        # Populate the major and minor schema versions from the URL in the describedBy field
        schema_semver = re.findall(r'\d+\.\d+\.\d+', metadata_doc["content"]["describedBy"])[0]
        provenance_core['schema_major_version'] = int(schema_semver.split(".")[0])
        provenance_core['schema_minor_version'] = int(schema_semver.split(".")[1])

        bundle_doc = metadata_doc['content']
        bundle_doc['provenance'] = provenance_core

        return bundle_doc

    def bundle_links(self, links):

        latest_schema = self.ingest_api.get_schemas(
            latest_only=True,
            high_level_entity='system',
            domain_entity='',
            concrete_entity='links'
        )

        schema_url = latest_schema[0]['_links']['json-schema']['href'] if latest_schema else None
        schema_version = latest_schema[0]['schemaVersion'] if latest_schema else None

        return {
            'describedBy': schema_url,
            'schema_type': 'link_bundle',
            'schema_version': schema_version,
            'links': links
        }

    def upload_metadata_files(self, submission_uuid, metadata_files_info):
        try:
            for metadata_type in ['project', 'biomaterial', 'process', 'protocol', 'file', 'links']:
                for metadata_doc in metadata_files_info[metadata_type]:
                    bundle_file = metadata_doc

                    if not bundle_file.get('is_from_input_bundle'):
                        metadata = MetadataResource.from_dict(bundle_file['original_doc'], require_provenance=False)
                        uploaded_file: StagingInfo = self.staging_service.stage_metadata(submission_uuid, metadata)
                        bundle_file['upload_file_url'] = uploaded_file.cloud_url
        except Exception as exception:
            message = "An error occurred on uploading bundle files: " + str(exception)
            raise BundleFileUploadError(message)

    # TODO handle error #export-errors
    def put_bundle_in_dss(self, bundle_uuid, bundle_version, created_files):
        try:
            created_bundle = self.dss_api.put_bundle(bundle_uuid, bundle_version, created_files)
        except BundleAlreadyExist as exception:
            raise BundleAlreadyExist(f"ERROR: Bundle already exists: {str(exception)}")
        except Exception as exception:
            raise BundleDSSError(f"ERROR: An error occurred while putting bundle in DSS: {str(exception)}")

        return created_bundle

    # TODO handle error #exporter-errors
    def put_files_in_dss(self, bundle_uuid, files_to_put, process_info):
        created_files = []

        for bundle_file in files_to_put:
            version = ''
            file_uuid = bundle_file["dss_uuid"]
            created_file = None
            input_data_files = [input_file['dataFileUuid'] for input_file in list(process_info.input_files.values())]

            try:
                # TODO if file is an input file, this file may already be in the data store, need to get the stored
                #  version
                # This assumes that the latest version is the file version in the input bundle, should be a safe
                # assumption for now
                # Ideally, bundle manifest must store the file uuid and version and version must be retrieved from there

                # if metadata file , check is_from_input_bundle flag, if true, do not put file to DSS again
                if bundle_file.get('is_from_input_bundle') or file_uuid in input_data_files:
                    file_response = self.dss_api.head_file(bundle_file["dss_uuid"])
                    created_file = {
                        'version': file_response.headers['X-DSS-VERSION']
                    }
                else:
                    created_file = self.dss_api.put_file(bundle_uuid, bundle_file)

                version = created_file['version']
            except Exception as exception:
                raise FileDSSError('An error occurred while putting file in DSS' + str(exception))

            file_param = {
                "indexed": bundle_file["indexed"],
                "name": bundle_file["submittedName"],
                "uuid": file_uuid,
                "content-type": bundle_file["content-type"],
                "version": version
            }

            created_files.append(file_param)

        return created_files

    def verify_files(self, created_files):
        for created_file in created_files:
            try:
                polling.poll(
                    lambda: self._is_file_copied(created_file),
                    step=30,
                    timeout=1200  # 20 minutes
                )
            except polling.TimeoutException:
                self.logger.error("File %s/%s with name %s takes too long to be copied.", created_file["uuid"],
                                  created_file["version"], created_file["name"])
                raise
            self.logger.info("File %s/%s with name %s is successfully copied!", created_file["uuid"],
                             created_file["version"], created_file["name"])

    def _is_file_copied(self, created_file):
        try:
            r = self.dss_api.head_file(created_file["uuid"], version=created_file["version"])
            return (r.status_code == requests.codes.ok) or (r.status_code == requests.codes.created)
        except Exception:
            return False

    def get_metadata_files(self, metadata_files_info):
        metadata_files = []

        for entity_type in ['biomaterial', 'file', 'project', 'protocol', 'process', 'links']:
            for metadata_file in metadata_files_info[entity_type]:
                metadata_files.append({
                    'name': metadata_file['upload_filename'],
                    'submittedName': metadata_file['dss_filename'],
                    'url': metadata_file.get('upload_file_url'),
                    'dss_uuid': metadata_file['dss_uuid'],
                    'indexed': metadata_file['indexed'],
                    'content-type': metadata_file['content_type'],
                    'update_date': metadata_file.get('update_date'),
                    'is_from_input_bundle': metadata_file.get('is_from_input_bundle')
                })
        return metadata_files

    def get_data_files(self, uuid_file_dict):
        data_files = []
        #  TODO: need to keep track of UUIDs used so that retries work when the DSS returns a 500
        for _, data_file in uuid_file_dict.items():
            filename = data_file['fileName']
            cloud_url = data_file['cloudUrl']
            data_file_uuid = data_file['dataFileUuid']

            data_files.append({
                'name': filename,
                'submittedName': filename,
                'url': cloud_url,
                'dss_uuid': data_file_uuid,
                'indexed': False,
                'content-type': 'data'
            })

        return data_files

    def create_bundle_manifest(self, bundle_manifest, files_by_type):
        bundle_manifest.fileProjectMap = dict()
        for metadata_file in files_by_type['project']:
            bundle_manifest.fileProjectMap[metadata_file['dss_uuid']] = [metadata_file['dss_uuid']]

        bundle_manifest.fileBiomaterialMap = dict()
        for metadata_file in files_by_type['biomaterial']:
            bundle_manifest.fileBiomaterialMap[metadata_file['dss_uuid']] = [metadata_file['dss_uuid']]

        bundle_manifest.fileProcessMap = dict()
        for metadata_file in files_by_type['process']:
            bundle_manifest.fileProcessMap[metadata_file['dss_uuid']] = [metadata_file['dss_uuid']]

        bundle_manifest.fileProtocolMap = dict()
        for metadata_file in files_by_type['protocol']:
            bundle_manifest.fileProtocolMap[metadata_file['dss_uuid']] = [metadata_file['dss_uuid']]

        bundle_manifest.fileFilesMap = dict()
        for metadata_file in files_by_type['file']:
            bundle_manifest.fileFilesMap[metadata_file['dss_uuid']] = [metadata_file['dss_uuid']]

        return bundle_manifest

    def get_concrete_entity_type(self, schema_uri):
        return schema_uri["content"]["describedBy"].rsplit('/', 1)[-1]

    def dump_to_file(self, content, filename, output_dir='output'):
        directory = os.path.abspath(output_dir)

        if not os.path.exists(directory):
            os.makedirs(directory)

        tmp_file = open(directory + "/" + filename + ".json", "w")
        tmp_file.write(content)
        tmp_file.close()


class File:
    def __init__(self):
        self.name = ""
        self.content_type = ""
        self.size = ""
        self.id = ""
        self.checksums = {}


class ProcessInfo:
    def __init__(self):
        self.project = {}

        # uuid => object mapping
        self.input_biomaterials = {}
        self.derived_by_processes = {}
        self.input_files = {}
        self.derived_files = {}
        self.protocols = {}
        self.supplementary_files = {}

        self.links = LinkSet()

        self.input_bundle = None


class LinkSet:
    def __init__(self):
        self.link_uuids = set()  # the uuid of a link is just the uuid of the process denoting the link
        self.links = []

    def add_link(self, link: dict):
        link_uuid = link["process"]
        if link_uuid in self.link_uuids:
            pass
        else:
            self.link_uuids.add(link_uuid)
            self.links.append(link)

    def get_links(self):
        return [deepcopy(link) for link in self.links]



class IngestError:
    """
    An error that can be reported to ingest.
    """

    def __init__(self):
        self.type = "http://ingest.data.humancellatlas.org/Error"
        self.title = ""
        self.detail = ""

    def getJSON(self):
        return self.__dict__


class ImporterError(IngestError):
    def __init__(self, detail=""):
        self.type = "http://importer.ingest.data.humancellatlas.org/Error"
        self.title = "An error occurred importing the file."
        self.detail = detail


class ParserError(ImporterError):
    def __init__(self, location="", type="", detail=""):
        self.type = "http://parser.importer.ingest.data.humancellatlas.org/Error"
        self.title = f'An error occurred parsing the file: {type}.'
        self.detail = f'{location}, error={detail}'


class SubmissionError(ImporterError):
    def __init__(self, detail=""):
        self.type = "http://submission.importer.ingest.data.humancellatlas.org/Error"
        self.title = "An error occurred submitting the file."
        self.detail = detail


class ExporterError(IngestError):
    def __init__(self, detail=""):
        self.type = "http://exporter.ingest.data.humancellatlas.org/Error"
        self.title = "Error occurred while attempting to export the submission."
        self.detail = detail



#!/usr/bin/env python
"""
Given a tabs template and list of schema URLs, will output a spreadsheet in Xls format
"""
__author__ = "jupp & hewgreen"
__license__ = "Apache 2.0"
__date__ = "31/01/2019"

import sys

from ingest.template.spreadsheet_builder import SpreadsheetBuilder

DEFAULT_INGEST_URL = "http://api.ingest.data.humancellatlas.org"
DEFAULT_SCHEMAS_ENDPOINT = "/schemas/search/latestSchemas"


class LinkedSpreadsheetBuilder(SpreadsheetBuilder):
    def __init__(self, output_file, hide_row=False, link_config=False, autofill_scale=0):
        super().create_initial_spreadsheet(output_file, hide_row)

        self.link_config = link_config
        self.autofill_scale = autofill_scale

    def build(self, spreadsheet_tabs_template):
        # TODO(maniarathi): Fill out docstring here once finished parsing and consolidating code.
        tab_config = spreadsheet_tabs_template.tab_config

        if (self.link_config is not False) and (self.autofill_scale != 0):  # some precalculation for whole sheet
            self._value_linking()

        for tab in tab_config.lookup("tabs"):

            for tab_name, detail in tab.items():

                if self.link_config is True:  # skip protocols and entities not in link_config
                    backbone = self.link_config[0]
                    len_check = [len(y) for y in backbone]
                    if all(x == len_check[0] for x in len_check):
                        self.backbone_entities = [list(y.keys())[0] for y in backbone]
                        self._protocol_linking(self.backbone_entities)
                    else:
                        print('Too many elements provided in backbone config. Aborting.')
                        sys.exit()

                    # todo project sub tabs contact etc need splitting out
                    always_add = ['project']  # todo NB hardcoded
                    if (tab_name not in self.backbone_entities) and (tab_name not in self.protocols_to_add) and (
                            tab_name not in always_add):
                        continue  # dont put the tab in if it isn't needed

                worksheet = self.spreadsheet.add_worksheet(detail["display_name"])

                for column_index, column_name in enumerate(detail["columns"]):

                    formatted_column_name = self.get_user_friendly_column_name(spreadsheet_tabs_template,
                                                                               column_name).upper()

                    if column_name.split(".")[-1] == "text":
                        desc = self.get_value_for_column(spreadsheet_tabs_template, column_name, "description")
                        required = bool(self.get_value_for_column(spreadsheet_tabs_template, column_name, "required"))
                        example_text = self.get_value_for_column(spreadsheet_tabs_template, column_name, "example")
                        guidelines = self.get_value_for_column(spreadsheet_tabs_template, column_name, "guidelines")
                    else:
                        if column_name + ".text" not in detail["columns"]:
                            desc = self.get_value_for_column(spreadsheet_tabs_template, column_name, "description")
                            required = bool(
                                self.get_value_for_column(spreadsheet_tabs_template, column_name, "required"))
                            example_text = self.get_value_for_column(spreadsheet_tabs_template, column_name, "example")
                            guidelines = self.get_value_for_column(spreadsheet_tabs_template, column_name, "guidelines")

                    if required:
                        formatted_column_name += " (Required)"

                    # set the user friendly name
                    worksheet.write(0, column_index, formatted_column_name, self.header_format)

                    column_width = len(formatted_column_name) if len(formatted_column_name) > 25 else 25

                    worksheet.set_column(column_index, column_index, column_width)

                    # set the description
                    worksheet.write(1, column_index, desc, self.desc_format)

                    # write example
                    worksheet.write(2, column_index,
                                    guidelines + ' For example: ' + example_text if example_text else guidelines,
                                    self.desc_format)

                    # set the key
                    worksheet.write(3, column_index, column_name, self.locked_format)

                    if column_name.split(".")[-1] == "ontology" or column_name.split(".")[-1] == "ontology_label":
                        worksheet.set_column(column_index, column_index, None, None, {'hidden': True})

                    if self.hidden_row:
                        worksheet.set_row(3, None, None, {'hidden': True})

                    if column_index == 0:
                        worksheet.set_row(0, 30)
                        worksheet.set_row(4, 30)

                        worksheet.write(4, column_index, "FILL OUT INFORMATION BELOW THIS ROW", self.header_format)

                    else:
                        worksheet.write(4, column_index, '', self.header_format)

                    # ADD EXAMPLES HERE
                    if self.autofill_scale > 0:  # flag for adding example info
                        self._fill_examples_from_schema(example_text, worksheet, column_name, column_index, tab_name)

                if self.link_config is not False:  # after normal column_name added to tab add linking column_name

                    self._make_col_name_mapping(spreadsheet_tabs_template)  # makes lookup dict for uf tab names
                    self._add_link_cols(tab_name, column_index, worksheet, self.header_format, self.backbone_entities)

                    # todo add process column if on seq tab

        if self.include_schemas_tab:
            self._write_schemas(spreadsheet_tabs_template.get_schema_urls())

    def _fill_examples_from_schema(self, example_text, worksheet, cols, col_number, tab_name):
        double_prefix = 'Should be one of: '
        metadata_fs = ';'
        if example_text.startswith(double_prefix):
            example_text_ = example_text[len(double_prefix):].split(',')[0]  # TODO hard coded assumption?
        elif metadata_fs in example_text:
            example_text_ = example_text.split(metadata_fs)[0]
        else:
            example_text_ = example_text

        prog_name = cols.split('.')

        one_liner_tabs = ['project']
        if (tab_name.endswith('protocol')) or (tab_name in one_liner_tabs):
            row_range_fill = 1
        else:
            row_range_fill = self.tab_multiplier.get(tab_name).get('y') * self.autofill_scale

        for x in range(row_range_fill):
            row_no = 5 + x

            # the first ID field
            if (len(prog_name) == 3) and (prog_name[2].endswith('_id') and (
                    prog_name[1].endswith('_core') and (col_number == 0))):  # TODO hard coded assumption?
                example_text_ = prog_name[0] + '_' + str(x)  # TODO add row counter here to support multiple row fills
            # the name field
            if (len(prog_name) == 3) and (
                    prog_name[2].endswith('_name') and (prog_name[1].endswith('_core'))):  # TODO hard coded assumption?
                example_text_ = prog_name[0] + '_' + str(x)
            # the description field
            if (len(prog_name) == 3) and (prog_name[2].endswith('_description') and (
                    prog_name[1].endswith('_core'))):  # TODO hard coded assumption?
                example_text_ = 'This is a description of ' + prog_name[0]

            worksheet.write(row_no, col_number, example_text_, self.locked_format)

            # project and protocol doesn't need these multiple rows

    def _value_linking(self):

        # # work out from backbone list and autofill_scale how many id values to fill
        # backbone = self.link_config[0]
        # multiplier = []
        # for entity_item in backbone:
        #     for entity,bundle_multiplier in entity_item.items(): # should only ever be one item checked elsewhere
        #         total_multiplier = bundle_multiplier * self.autofill_scale
        #         multiplier_ = {}
        #         multiplier_[entity] = total_multiplier
        #         multiplier.append(multiplier_) # list to maintain backbone ordering
        #
        # self.multiplier = multiplier

        backbone = self.link_config[0]
        pre_multiplier = []
        for entity in backbone:
            entity_entry = {}
            prev_entity = None
            for tab_name, y in entity.items():
                entity_entry['tab_name'] = tab_name
                entity_entry['y'] = y
                if len(pre_multiplier):
                    entity_entry['text'] = prev_entity
                else:
                    entity_entry['text'] = None
                prev_entity = tab_name
                pre_multiplier.append(entity_entry)

        post_multiplier = []
        backbone_index = 0
        for sheet in pre_multiplier:
            mid_multiplier = []
            counter = 0
            if backbone_index != 0:
                y = sheet.get('y') * self.autofill_scale
                prev_y = pre_multiplier[backbone_index - 1].get('y') * self.autofill_scale
                if y >= prev_y:
                    print_count = int(y / prev_y)
                    for row in range(prev_y):
                        for row_ in range(print_count):
                            if sheet.get('text'):
                                to_print = sheet.get('text') + '_' + str(counter)
                            else:
                                to_print = '_' + str(counter)
                            mid_multiplier.append(to_print)
                        counter += 1
                else:
                    print_count = int(prev_y / y)
                    counter = 0
                    for row in range(y):
                        to_print = ''

                        for row_ in range(print_count):
                            to_print += sheet.get('text') + '_' + str(counter) + '||'
                            counter += 1
                        mid_multiplier.append(to_print[:-2])
                sheet['pre_comb_linking'] = mid_multiplier
            post_multiplier.append(sheet)

            backbone_index += 1

        uniques = []
        tab_multiplier = {}
        for entity in post_multiplier:
            if entity.get('tab_name') not in uniques:
                uniques.append(entity.get('tab_name'))
                tab_multiplier[entity.get('tab_name')] = entity  # add pop tab name
            else:
                prev_entity = tab_multiplier.get(entity.get('tab_name'))
                prev_entity['y'] = entity.get('y') + prev_entity.get('y')
                prev_entity['pre_comb_linking'] = prev_entity.get('pre_comb_linking') + entity.get('pre_comb_linking')

        self.tab_multiplier = tab_multiplier

    def _add_link_cols(self, tab_name, col_number, worksheet, hf, backbone_entities):

        # given the tab name work out what entity links need to be added

        _link_to_tab = []
        index_list = [i for i, e in enumerate(backbone_entities) if e == tab_name]
        for x in index_list:
            if x > 0:
                _link_to_tab.append(backbone_entities[x - 1])

        # add link columns
        for link_to_tab in _link_to_tab:
            display_name = self.col_name_mapping.get(link_to_tab)[0]
            prog_name = self.col_name_mapping.get(link_to_tab)[1]
            uf = str('DERIVED FROM {}'.format(display_name.upper()))
            desc = str('Enter biomaterial ID from "{}" tab that this entity was derived from.'.format(display_name))

            col_number = self._write_column_head(worksheet, col_number, uf, hf, desc, prog_name, tab_name)

        # add protocol columns
        # all_to_add = []
        reverse_dict = {}
        for k, v in self.protocols_to_add.items():
            # all_to_add += v
            try:
                for entity in v:
                    if entity in reverse_dict:
                        reverse_dict[entity] = reverse_dict.get(entity) + [k]
                    else:
                        reverse_dict[entity] = [k]
            except TypeError:
                continue

        if tab_name in reverse_dict:
            add_these = reverse_dict.get(tab_name)
            for protocol in add_these:
                display_name = self.col_name_mapping.get(protocol)[0]
                prog_name = self.col_name_mapping.get(protocol)[1]
                uf = str('ID OF {} USED'.format(display_name.upper()))
                desc = str('Enter protocol ID from "{}" tab that this entity was derrived from.'.format(display_name))
                col_number = self._write_column_head(worksheet, col_number, uf, hf, desc, prog_name, tab_name)

    def _write_column_head(self, worksheet, col_number, uf, hf, desc, prog_name, tab_name):
        worksheet.write(0, col_number, uf, hf)  # user friendly name
        worksheet.write(1, col_number, desc, self.desc_format)  # description
        # worksheet.write(2, col_number, ???, self.desc_format) # example
        worksheet.write(3, col_number, prog_name, self.locked_format)  # programatic name
        worksheet.write(4, col_number, '', hf)  # blank column

        # todo loop to fill in values for links (this func only fill link back
        # AKA ADD EXAMPLES

        if (self.link_config is not False) and (self.autofill_scale != 0):
            prog_name_list = prog_name.split('.')
            link_fill = self.tab_multiplier.get(tab_name).get('pre_comb_linking')
            row_no = 5
            if (len(prog_name_list) == 3) and (prog_name_list[2].endswith('biomaterial_id') and (
                    prog_name_list[1].endswith('_core'))):
                for x in link_fill:
                    worksheet.write(row_no, col_number, x, self.locked_format)
                    row_no += 1
            else:
                fill = prog_name_list[0] + str('_0')
                for x in range(len(link_fill)):
                    worksheet.write(row_no, col_number, fill, self.locked_format)
                    row_no += 1
        col_number += 1
        return col_number

    def _protocol_linking(self, backbone_entities):
        protocol_pairings = self.link_config[1]
        # protocols_to_add = []
        protocols_to_add = {}

        index_counter = 0
        for entity in backbone_entities:
            index_counter += 1
            the_source = entity
            try:
                the_output = backbone_entities[index_counter]
            except IndexError:
                self.protocols_to_add = protocols_to_add
                return
            for key, value in protocol_pairings.items():
                protocol_prog_name = key
                for pairing in value:
                    if (pairing.get('source') == the_source) and (pairing.get('output') == the_output):
                        # protocols_to_add.append(protocol_prog_name)
                        if protocol_prog_name in protocols_to_add:
                            protocols_to_add[protocol_prog_name] = protocols_to_add.get(protocol_prog_name).append(
                                the_output)
                        else:
                            protocols_to_add[protocol_prog_name] = [the_output]

    def _make_col_name_mapping(self, template):
        col_name_mapping = {}
        for entity_type in template._template.get('tabs'):
            for key, value in entity_type.items():
                display_name = value.get('display_name')
                for col in value.get('columns'):
                    dot_parse = col.split('.')
                    if (len(dot_parse) == 3) and (
                            (dot_parse[2] == 'protocol_id') or (dot_parse[2] == 'biomaterial_id')) and (
                            dot_parse[1].endswith('_core')):  # todo WARNING verging on hard coded
                        prog_name = col

                col_name_mapping[key] = [display_name, prog_name]
        self.col_name_mapping = col_name_mapping



#!/usr/bin/env python
"""
desc goes here
"""
import json
import logging
import os
import time
from urllib.parse import urljoin, quote

import requests

from ingest.api.requests_utils import create_session_with_retry


class IngestApi:
    def __init__(self, url=None, token_manager=None):
        self.logger = logging.getLogger(__name__)
        self.session = create_session_with_retry()
        self.token_manager = token_manager

        if not url and 'INGEST_API' in os.environ:
            url = os.environ['INGEST_API']
            # expand interpolated env vars
            url = os.path.expandvars(url)
        self.url = url if url else "http://localhost:8080"
        self.headers = {'Content-type': 'application/json'}
        self.token = None
        self._submission_links = {}
        self.logger.info(f"using {self.url} for ingest API")

        self._ingest_links = self._get_ingest_links()

    def get(self, url, **kwargs):
        if 'headers' not in kwargs:
            kwargs['headers'] = {'Content-type': 'application/json'}
        return self.session.get(url, **kwargs)

    def patch(self, url, patch):
        r = self.session.patch(url, json=patch, headers=self.get_headers())
        r.raise_for_status()
        return r

    def put(self, url, data):
        r = self.session.put(url, json=data, headers=self.get_headers())
        return r

    def set_token(self, token):
        self.token = token
        self.headers['Authorization'] = self.token
        self.logger.debug(f'Token set!')

        return self.headers

    def get_headers(self):
        # refresh token
        if self.token_manager:
            self.set_token(f'Bearer {self.token_manager.get_token()}')
            self.logger.debug(f'Token refreshed!')

        return self.headers

    def _get_ingest_links(self):
        r = self.get(self.url, headers=self.get_headers())
        r.raise_for_status()
        return r.json()["_links"]

    def get_link_from_resource_url(self, resource_url, link_name):
        r = self.get(resource_url, headers=self.get_headers())
        r.raise_for_status()
        links = r.json().get('_links', {})
        return links.get(link_name, {}).get('href')

    def get_link_from_resource(self, resource, link_name):
        links = resource.get('_links', {})
        return links.get(link_name, {}).get('href')

    def get_schemas(self, latest_only=True, high_level_entity=None, domain_entity=None, concrete_entity=None):
        schema_url = self.get_schemas_url()
        all_schemas = []

        if latest_only:
            search_url = self.get_link_from_resource_url(schema_url, "search")
            r = self.get(search_url, headers=self.get_headers())
            if r.status_code == requests.codes.ok:
                response_j = json.loads(r.text)
                all_schemas = list(self.get_related_entities("latestSchemas", response_j, "schemas"))
        else:
            all_schemas = list(self.get_entities(schema_url, "schemas"))

        if high_level_entity:
            all_schemas = list(filter(lambda schema: schema.get('highLevelEntity') == high_level_entity, all_schemas))

        if domain_entity:
            all_schemas = list(filter(lambda schema: schema.get('domainEntity') == domain_entity, all_schemas))

        if concrete_entity:
            all_schemas = list(filter(lambda schema: schema.get('concreteEntity') == concrete_entity, all_schemas))

        return all_schemas

    def get_schemas_url(self):
        return self.get_resource_repository_url("schemas")

    def get_staging_jobs_url(self):
        return self.get_resource_repository_url("stagingJobs")

    def get_resource_repository_url(self, repository_name: str):
        if repository_name in self._ingest_links:
            return self._ingest_links[repository_name]["href"].rsplit("{")[0]
        return None

    def getSubmissions(self):
        params = {'sort': 'submissionDate,desc'}
        r = self.get(self._ingest_links["submissionEnvelopes"]["href"].rsplit("{")[0], params=params,
                     headers=self.get_headers())
        if r.status_code == requests.codes.ok:
            return json.loads(r.text)["_embedded"]["submissionEnvelopes"]

    def getProjects(self, id):
        submissionUrl = self.url + '/submissionEnvelopes/' + id + '/projects'
        r = self.get(submissionUrl, headers=self.get_headers())
        projects = []
        if r.status_code == requests.codes.ok:
            projects = json.loads(r.text)
        return projects

    def get_project_by_id(self, id):
        submission_url = self.url + '/projects/' + id
        r = self.get(submission_url, headers=self.get_headers())
        if r.status_code == requests.codes.ok:
            project = json.loads(r.text)
            return project
        else:
            raise ValueError("Project " + id + " could not be retrieved")

    def get_project_by_uuid(self, uuid):
        return self.get_entity_by_uuid('projects', uuid)

    def get_entity_by_uuid(self, entity_type, uuid):
        url = self.url + f'/{entity_type}/search/findByUuid?uuid=' + uuid

        # TODO make the endpoint consistent
        if entity_type == 'submissionEnvelopes':
            url = self.url + f'/{entity_type}/search/findByUuidUuid?uuid=' + uuid

        r = self.get(url, headers=self.get_headers())
        r.raise_for_status()
        return r.json()

    def get_entity_by_callback_link(self, callback_link):
        url = f'{self.url}{callback_link}'
        r = self.get(url, headers=self.get_headers())
        r.raise_for_status()
        return r.json()

    def get_file_by_submission_url_and_filename(self, submission_url, filename):
        search_url = self.get_link_from_resource_url(self.url + '/files/search',
                                                     'findBySubmissionEnvelopesInAndFileName')
        search_url = search_url.replace('{?submissionEnvelope,fileName}', '')
        r = self.get(search_url, params={'submissionEnvelope': submission_url, 'fileName': filename})
        if r.status_code == requests.codes.ok:
            return r.json()
        return None

    def get_submission(self, submission_url):
        r = self.get(submission_url, headers=self.get_headers())
        r.raise_for_status()
        return r.json()

    def get_submission_by_uuid(self, submission_uuid):
        search_link = self.get_link_from_resource_url(self.url + '/submissionEnvelopes/search', 'findByUuid')
        search_link = search_link.replace('{?uuid}', '')  # TODO: use a REST traverser instead of requests?
        r = self.get(search_link, params={'uuid': submission_uuid})
        r.raise_for_status()
        return r.json()

    def get_files(self, id):
        submission_url = self.url + '/submissionEnvelopes/' + id + '/files'
        r = self.get(submission_url, headers=self.get_headers())
        files = []
        if r.status_code == requests.codes.ok:
            files = json.loads(r.text)
        return files

    def get_bundle_manifests(self, id):
        submission_url = self.url + '/submissionEnvelopes/' + id
        return self.get_entities(submission_url, "bundleManifests", 500)

    def create_submission(self, update_submission=False):
        try:
            create_submission_url = self._ingest_links["submissionEnvelopes"]["href"].rsplit("{")[0]

            if update_submission:
                create_submission_url = f'{create_submission_url}/updateSubmissions'

            r = self.session.post(create_submission_url, data="{}",
                                  headers=self.get_headers())
            r.raise_for_status()
            submission = r.json()
            submission_url = submission["_links"]["self"]["href"].rsplit("{")[0]
            self._submission_links[submission_url] = submission["_links"]
            return submission
        except requests.exceptions.RequestException as err:
            self.logger.error("Request failed: " + str(err))
            raise

    def get_submission_links(self, submission_url):
        if not self._submission_links.get(submission_url):
            r = self.get(submission_url, headers=self.get_headers())
            r.raise_for_status()
            self._submission_links[submission_url] = r.json()["_links"]

        return self._submission_links.get(submission_url)

    def get_link_in_submission(self, submission_url, link_name):
        links = self.get_submission_links(submission_url)
        if link_name in links:
            link_obj = links.get(link_name)
            link = link_obj['href'].rsplit("{")[0]
            return link

        raise ValueError(f"{link_name} is not in submission resource links")

    def update_submission_state(self, submission_id, state):
        state_url = self.get_submission_state_url(submission_id, state)
        r = self.session.put(state_url, headers=self.get_headers())
        r.raise_for_status()
        return r.json()

    def get_submission_state_url(self, submission_id, state):
        submission_url = self.get_submission_url(submission_id)
        return self.get_link_in_submission(submission_url, state)

    def getSubmissionUri(self, submission_id):
        return self._ingest_links["submissionEnvelopes"]["href"].rsplit("{")[0] + "/" + submission_id

    def get_submission_url(self, submission_id):
        return self._ingest_links["submissionEnvelopes"]["href"].rsplit("{")[0] + "/" + submission_id

    def get_full_url(self, callback_link):
        return urljoin(self.url, callback_link)

    def get_process(self, process_url):
        r = self.get(process_url, headers=self.get_headers())
        r.raise_for_status()
        return r.json()

    def get_entities(self, submission_url, entity_type):
        r = self.get(submission_url, headers=self.get_headers())
        if r.status_code == requests.codes.ok:
            if entity_type in json.loads(r.text)["_links"]:
                yield from self.get_all(json.loads(r.text)["_links"][entity_type]["href"], entity_type)

    def get_all(self, url, entity_type):
        r = self.get(url, headers=self.get_headers())
        r.raise_for_status()
        result = r.json()

        entities = result["_embedded"][entity_type] if '_embedded' in result else []
        yield from entities

        while "next" in result["_links"]:
            next_url = result["_links"]["next"]["href"]
            r = self.get(next_url, headers=self.get_headers())
            r.raise_for_status()
            result = r.json()
            entities = result["_embedded"][entity_type]
            yield from entities
            self.logger.info(f"GET {entity_type} {json.dumps(result['page'])}")

    def get_related_entities(self, relation, entity, entity_type):
        # get the self link from entity
        if relation in entity["_links"]:
            entity_uri = entity["_links"][relation]["href"]
            for entity in self.get_all(entity_uri, entity_type):
                yield entity

    def get_related_entities_count(self, relation, entity, entity_type):
        if relation in entity["_links"]:
            entity_uri = entity["_links"][relation]["href"]
            r = self.get(entity_uri)
            r.raise_for_status()
            result = r.json()
            page = result.get('page')
            if page:
                return page.get('totalElements')
            return len(result["_embedded"][entity_type])

    def create_project(self, submission_url, content, uuid=None):
        return self.create_entity(submission_url, content, "projects", uuid)

    def create_biomaterial(self, submission_url, content, uuid=None):
        return self.create_entity(submission_url, content, "biomaterials", uuid)

    def create_process(self, submission_url, json_object, uuid=None):
        return self.create_entity(submission_url, json_object, "processes", uuid)

    def create_protocol(self, submission_url, content, uuid=None):
        return self.create_entity(submission_url, content, "protocols", uuid)

    def create_file(self, submission_url, filename, content, uuid=None):
        submission_files_url = self.get_link_in_submission(submission_url, 'files')

        submission_files_url = submission_files_url + "/" + quote(filename)

        file_to_create_object = {
            "fileName": filename,
            "content": content
        }

        params = {}
        if uuid:
            params["updatingUuid"] = uuid

        time.sleep(0.001)

        # Do not use session retries here as it will throw requests.exceptions.RetryError for http 409 error
        # We want to retry that error when creating files by doing a subsequent PATCH requests to update the metadata
        r = requests.post(submission_files_url, json=file_to_create_object,
                          headers=self.get_headers(), params=params)

        # TODO Investigate why core is returning internal server error
        if r.status_code == requests.codes.conflict or r.status_code == requests.codes.internal_server_error:
            search_files = self.get_file_by_submission_url_and_filename(submission_url, filename)

            if search_files and search_files.get('_embedded') and search_files['_embedded'].get('files'):
                file_in_ingest = search_files['_embedded'].get('files')[0]
                existing_content = file_in_ingest.get('content')
                new_content = existing_content

                if existing_content:
                    new_content.update(content)
                else:
                    new_content = content

                file_url = file_in_ingest['_links']['self']['href']
                time.sleep(0.001)
                r = self.session.patch(file_url, json={'content': new_content}, headers=self.get_headers())
                self.logger.debug(f'Updating existing content of file {file_url}.')

        r.raise_for_status()

        return r.json()

    def create_submission_manifest(self, submission_url, data):
        return self.create_entity(submission_url, data, 'submissionManifest')

    def create_submission_error(self, submission_url, data):
        return self.create_entity(submission_url, data, 'submissionEnvelopeErrors')

    def create_entity(self, submission_url, content, entity_type, uuid=None):
        params = {}
        if uuid:
            params["updatingUuid"] = uuid

        submission_url = self.get_link_in_submission(submission_url, entity_type)
        self.logger.debug(f"POST {submission_url} {json.dumps(content)}")

        r = self.session.post(submission_url, json=content, headers=self.get_headers(), params=params)
        r.raise_for_status()
        return r.json()

    def get_object_uuid(self, entity_uri):
        r = self.get(entity_uri, headers=self.get_headers())
        if r.status_code == requests.codes.ok:
            return json.loads(r.text)["uuid"]["uuid"]

    def link_entity(self, from_entity, to_entity, relationship):
        if not from_entity:
            raise ValueError("Error: from_entity is None")

        if not to_entity:
            raise ValueError("Error: to_entity is None")

        if not relationship:
            raise ValueError("Error: relationship is None")

        # check each dict in turn for non-None-ness

        from_entity_links = from_entity["_links"] if "_links" in from_entity else None
        if not from_entity_links:
            raise ValueError("Error: from_entity has no _links")

        from_entity_links_relationship = from_entity_links[relationship] if relationship in from_entity_links else None
        if not from_entity_links_relationship:
            raise ValueError("Error: from_entity_links has no {0} relationship".format(relationship))

        from_entity_links_relationship_href = from_entity_links_relationship[
            "href"] if "href" in from_entity_links_relationship else None
        if not from_entity_links_relationship_href:
            raise ValueError("Error: from_entity_links_relationship for relationship {0} has no href".format(relationship))

        from_uri = from_entity["_links"][relationship]["href"]
        to_uri = self.get_link_from_resource(to_entity, 'self')

        self.logger.info('fromUri ' + from_uri + ' toUri:' + to_uri)

        headers = {
            'Content-type': 'text/uri-list',
            'Authorization': self.get_headers()['Authorization']
        }
        r = self.session.post(from_uri.rsplit("{")[0],
                              data=to_uri.rsplit("{")[0], headers=headers)
        r.raise_for_status()

        return r

    def create_bundle_manifest(self, bundleManifest):
        url = self._ingest_links["bundleManifests"]["href"].rsplit("{")[0]
        r = self.session.post(url, json=bundleManifest.__dict__, headers=self.get_headers())
        r.raise_for_status()
        return r.json()

    def update_staging_details(self, submission_url, uuid, staging_area_location):
        staging_details = {
            "stagingDetails": {
                "stagingAreaUuid": {
                    "uuid": uuid
                },
                "stagingAreaLocation": {
                    "value": staging_area_location
                }
            }
        }
        r = self.session.patch(submission_url, data=json.dumps(staging_details))
        r.raise_for_status()

    def delete_staging_jobs(self, staging_area_uuid):
        delete_jobs_url = f'{self.get_staging_jobs_url()}/delete'
        r = self.session.delete(delete_jobs_url, params={"stagingAreaUuid": staging_area_uuid},
                                headers=self.get_headers())
        r.raise_for_status()
        return r.json()

    def delete_staging_job(self, staging_job_url):
        r = self.session.delete(staging_job_url, headers=self.get_headers())
        r.raise_for_status()
        return r

    def create_staging_job(self, staging_area_uuid, file_name, metadata_uuid):
        staging_info = {
            "stagingAreaUuid": staging_area_uuid,
            "stagingAreaFileName": file_name,
            "metadataUuid": metadata_uuid
        }
        r = requests.post(self.get_staging_jobs_url(), json=staging_info, headers=self.get_headers())
        r.raise_for_status()
        return r.json()

    def complete_staging_job(self, complete_url, upload_file_uri):
        staging_info = {
            "stagingAreaFileUri": upload_file_uri
        }
        r = self.session.patch(complete_url, json=staging_info, headers=self.get_headers())
        r.raise_for_status()
        return r.json()

    def find_staging_job(self, upload_area_uuid, filename):
        search_params = {
            "stagingAreaUuid": upload_area_uuid,
            "stagingAreaFileName": filename
        }
        find_staging_job_url = f'{self.get_staging_jobs_url()}/search/findByStagingAreaUuidAndStagingAreaFileName'
        r = self.session.get(find_staging_job_url, params=search_params)
        r.raise_for_status()
        return r.json()



import logging

from ingest.importer.conversion import template_manager
from ingest.importer.conversion.metadata_entity import MetadataEntity
from ingest.importer.conversion.template_manager import TemplateManager
from ingest.importer.spreadsheet.ingest_workbook import IngestWorkbook
from ingest.importer.spreadsheet.ingest_worksheet import IngestWorksheet
from ingest.importer.submission import IngestSubmitter, EntityMap, EntityLinker, Submission
from ingest.template.exceptions import UnknownKeySchemaException
from ingest.utils.IngestError import ImporterError, ParserError

format = '[%(filename)s:%(lineno)s - %(funcName)20s() ] %(asctime)s - %(name)s - %(levelname)s - %(message)s'
logging.basicConfig(format=format)


class XlsImporter:
    """
    XlsImporter is used to convert a contributor's spreadsheet into metadata json entities and to submit those to
    Ingest. Please see https://github.com/HumanCellAtlas/ingest-central/wiki/Data-Contributors-Spreadsheet-Quick-Guide
    for more information on the spreadsheet format.
    """

    def __init__(self, ingest_api):
        self.ingest_api = ingest_api
        self.logger = logging.getLogger(__name__)

    def generate_json(self, file_path, project_uuid=None):
        ingest_workbook = IngestWorkbook.from_file(file_path)

        try:
            template_mgr = template_manager.build(ingest_workbook.get_schemas(), self.ingest_api)
        except Exception as e:
            raise SchemaRetrievalError(
                f'There was an error retrieving the schema information to process the spreadsheet. {str(e)}')

        workbook_importer = WorkbookImporter(template_mgr)
        spreadsheet_json, errors = workbook_importer.do_import(ingest_workbook, project_uuid)

        return spreadsheet_json, template_mgr, errors

    def dry_run_import_file(self, file_path, project_uuid=None):
        spreadsheet_json, template_mgr, errors = self.generate_json(file_path, project_uuid)

        if errors:
            return None, errors

        entity_map = self._process_links_from_spreadsheet(template_mgr, spreadsheet_json)

        return entity_map, []

    def import_file(self, file_path, submission_url, project_uuid=None):
        submission = None
        try:
            spreadsheet_json, template_mgr, errors = self.generate_json(file_path, project_uuid)
            if not errors:
                entity_map = self._process_links_from_spreadsheet(template_mgr, spreadsheet_json)

                # TODO the submission_url should be passed to the IngestSubmitter instead
                submitter = IngestSubmitter(self.ingest_api)
                submission = submitter.submit(entity_map, submission_url)
                self.logger.info(f'Submission in {submission_url} is done!')
            else:
                self.report_errors(submission_url, errors)
        except Exception as e:
            self.ingest_api.create_submission_error(submission_url, ImporterError(str(e)).getJSON())
            self.logger.error(str(e), exc_info=True)
        else:
            return submission, template_mgr

    def report_errors(self, submission_url, errors):
        self.logger.info(f'Logged {len(errors)} ParsingErrors.', exc_info=False)
        for error in errors:
            self.ingest_api.create_submission_error(
                submission_url,
                ParserError(error["location"], error["type"], error["detail"]).getJSON()
            )

    @staticmethod
    def _process_links_from_spreadsheet(template_mgr, spreadsheet_json):
        entity_map = EntityMap.load(spreadsheet_json)
        entity_linker = EntityLinker(template_mgr)
        entity_map = entity_linker.process_links_from_spreadsheet(entity_map)
        return entity_map

    @staticmethod
    def create_update_spreadsheet(submission: Submission, template_mgr: TemplateManager, file_path):
        if not submission:
            return
        wb = IngestWorkbook.from_file(file_path, read_only=False)
        wb.add_entity_uuids(submission)
        wb.add_schemas_worksheet(template_mgr.get_schemas())
        return wb.save(file_path)


_PROJECT_ID = 'project_0'
_PROJECT_TYPE = 'project'


class _ImportRegistry:
    """
    This is a helper class for managing metadata entities during Workbook import.
    """

    def __init__(self, template_mgr):
        self.template_mgr = template_mgr
        self._submittable_registry = {}
        self._module_registry = {}
        self._module_list = []
        self.project_id = _PROJECT_ID

    def add_submittable(self, metadata: MetadataEntity):
        # TODO no test to check case sensitivity
        domain_type = metadata.domain_type.lower()
        type_map = self._submittable_registry.get(domain_type)
        if not type_map:
            type_map = {}
            self._submittable_registry[domain_type] = type_map
        if domain_type.lower() == _PROJECT_TYPE:
            if not type_map.get(self.project_id):
                metadata.object_id = metadata.object_id or self.project_id
                self.project_id = metadata.object_id
            else:
                raise MultipleProjectsFound()
        type_map[metadata.object_id] = metadata

    def add_submittables(self, metadata_entities):
        for entity in metadata_entities:
            self.add_submittable(entity)

    def add_module(self, metadata: MetadataEntity):
        if metadata.domain_type.lower() == 'project':
            metadata.object_id = self.project_id
        self._module_list.append(metadata)

    def add_modules(self, module_field_name, metadata_entities):
        allowed_fields = [module_field_name]
        allowed_fields.extend(self.template_mgr.default_keys)
        removed_fields = []
        for entity in metadata_entities:
            removed_fields.extend(entity.list_fields(excluded_fields=allowed_fields))
            entity.retain_fields(module_field_name)
            self.add_module(entity)
        return removed_fields

    def import_modules(self):
        for module_entity in self._module_list:
            type_map = self._submittable_registry.get(module_entity.domain_type)
            submittable_entity = type_map.get(module_entity.object_id)
            submittable_entity.add_module_entity(module_entity)

    def flatten(self):
        flat_map = {}
        for domain_type, type_map in self._submittable_registry.items():
            flat_type_map = {object_id: metadata.map_for_submission()
                             for object_id, metadata in type_map.items()}
            flat_map[domain_type] = flat_type_map
        return flat_map

    def has_project(self):
        project_registry = self._submittable_registry.get(_PROJECT_TYPE)
        return project_registry and project_registry.get(self.project_id)


class WorkbookImporter:
    def __init__(self, template_mgr):
        self.worksheet_importer = WorksheetImporter(template_mgr)
        self.template_mgr = template_mgr
        self.logger = logging.getLogger(__name__)

    def do_import(self, workbook: IngestWorkbook, project_uuid=None):
        registry = _ImportRegistry(self.template_mgr)
        importable_worksheets = workbook.importable_worksheets()

        if project_uuid:
            project_metadata = MetadataEntity(domain_type=_PROJECT_TYPE,
                                              concrete_type=_PROJECT_TYPE,
                                              object_id=project_uuid,
                                              is_reference=True,
                                              content={})
            registry.add_submittable(project_metadata)

            importable_worksheets = [ws for ws in importable_worksheets
                                     if _PROJECT_TYPE not in ws.title.lower()]
        workbook_errors = []
        for worksheet in importable_worksheets:
            try:
                self.sheet_in_schemas(worksheet)
                metadata_entities, worksheet_errors = self.worksheet_importer.do_import(worksheet)
                module_field_name = worksheet.get_module_field_name()
                workbook_errors.extend(worksheet_errors)

                if worksheet.is_module_tab():
                    removed_data = registry.add_modules(module_field_name, metadata_entities)
                    workbook_errors.extend(self.list_data_removal_errors(worksheet.title, removed_data))
                else:
                    registry.add_submittables(metadata_entities)
            except Exception as e:
                workbook_errors.append(
                    {"location": f'sheet={worksheet.title}', "type": e.__class__.__name__, "detail": str(e)})

        if registry.has_project():
            registry.import_modules()
        else:
            e = NoProjectFound()
            workbook_errors.append({"location": "File", "type": e.__class__.__name__, "detail": str(e)})
        return registry.flatten(), workbook_errors

    def sheet_in_schemas(self, worksheet):
        schemas = self.template_mgr.template.json_schemas
        try:
            concrete_type = self.template_mgr.get_concrete_type(worksheet.title)
        except UnknownKeySchemaException as e:
            raise SheetNotFoundInSchemas(worksheet.title)
        module_field_name = worksheet.get_module_field_name()
        for schema in schemas:
            if 'name' in schema or 'title' in schema:
                schema_name = schema['name'] if 'name' in schema else schema['title']
                if schema_name == concrete_type:
                    if not worksheet.is_module_tab() or module_field_name in schema['properties']:
                        return True
                    raise SheetNotFoundInSchemas(worksheet.title)
        raise SheetNotFoundInSchemas(worksheet.title)

    @staticmethod
    def list_data_removal_errors(sheet, removed_data):
        errors = []
        for data in removed_data:
            e = DataRemoval(data['key'], data['value'])
            errors.append({"location": f'sheet={sheet}', "type": e.__class__.__name__, "detail": str(e)})
        return errors


class WorksheetImporter:
    KEY_HEADER_ROW_IDX = 4
    USER_FRIENDLY_HEADER_ROW_IDX = 2
    START_ROW_OFFSET = 5

    UNKNOWN_ID_PREFIX = '_unknown_'

    def __init__(self, template: TemplateManager):
        self.template = template
        self.unknown_id_ctr = 0
        self.logger = logging.getLogger(__name__)
        self.concrete_entity = None

    def do_import(self, ingest_worksheet: IngestWorksheet):
        records = []
        worksheet_errors = []
        try:
            row_template = self.template.create_row_template(ingest_worksheet)
            rows = ingest_worksheet.get_data_rows()
            for index, row in enumerate(rows):
                metadata, row_errors = row_template.do_import(row)
                for error in row_errors:
                    if 'location' in error:
                        error["location"] = f'sheet={ingest_worksheet.title} row={index}, {error["location"]}'
                    else:
                        error["location"] = f'sheet={ingest_worksheet.title} row={index}'
                    worksheet_errors.append(error)
                if not metadata.object_id:
                    metadata.object_id = self._generate_id()
                records.append(metadata)
        except Exception as e:
            worksheet_errors.append({
                "location": f'sheet={ingest_worksheet.title}',
                "type": e.__class__.__name__,
                "detail": str(e)
            })
        return records, worksheet_errors

    def _generate_id(self):
        self.unknown_id_ctr = self.unknown_id_ctr + 1
        return f'{self.UNKNOWN_ID_PREFIX}{self.unknown_id_ctr}'


class MultipleProjectsFound(Exception):
    def __init__(self):
        message = f'The spreadsheet should only be associated to a single project.'
        super(MultipleProjectsFound, self).__init__(message)


class NoProjectFound(Exception):
    def __init__(self):
        message = f'The spreadsheet should be associated to a project.'
        super(NoProjectFound, self).__init__(message)


class SheetNotFoundInSchemas(Exception):
    def __init__(self, sheet):
        message = f'The sheet named {sheet} was not found in the schema list.'
        super(SheetNotFoundInSchemas, self).__init__(message)
        self.sheet = sheet


class DataRemoval(Exception):
    def __init__(self, key, value):
        message = f'The column header [{key}] was not recognised, the following data has been removed: {value}.'
        super(DataRemoval, self).__init__(message)
        self.key = key
        self.value = value


class SchemaRetrievalError(Exception):
    pass



from ingest.api.ingestapi import IngestApi
from ingest.exporter.bundle import BundleService, Bundle
from ingest.exporter.metadata import MetadataService
from ingest.exporter.staging import StagingService


class SubmissionEnvelopeParseException(Exception):
    pass


class SubmissionEnvelope:

    def __init__(self, uuid, staging_area_uuid):
        self.uuid = uuid
        self.staging_area_uuid = staging_area_uuid

    @staticmethod
    def from_dict(source: dict):
        try:
            uuid = source['uuid']['uuid']
            staging_area_uuid = source['stagingDetails']['stagingAreaUuid']['uuid']
            return SubmissionEnvelope(uuid, staging_area_uuid)
        except (KeyError, TypeError) as e:
            raise SubmissionEnvelopeParseException(e)


class Exporter:

    def __init__(self, ingest_api: IngestApi, metadata_service: MetadataService,
                 staging_service: StagingService, bundle_service: BundleService):
        self.ingest_api = ingest_api
        self.metadata_service = metadata_service
        self.staging_service = staging_service
        self.bundle_service = bundle_service

    def export_update(self, submission_source: dict, bundle_uuid: str, metadata_urls: list,
                      update_version: str):
        bundle = self.bundle_service.fetch(bundle_uuid)
        submission = SubmissionEnvelope.from_dict(submission_source)
        staging_details = self._apply_metadata_updates(bundle, metadata_urls,
                                                       submission.staging_area_uuid)
        bundle.update_version(update_version)
        self.bundle_service.update(bundle, staging_details)
        manifest = bundle.generate_manifest(submission.uuid)
        self.ingest_api.create_bundle_manifest(manifest)

    def _apply_metadata_updates(self, bundle: Bundle, metadata_urls, staging_area_uuid):
        staging_details = []
        for url in metadata_urls:
            metadata_resource = self.metadata_service.fetch_resource(url)
            staging_info = self.staging_service.stage_metadata(staging_area_uuid, metadata_resource)
            staging_details.append(staging_info)
            bundle.update_file(metadata_resource)
        return staging_details



import logging
from enum import Enum

from ingest.importer.conversion import utils, data_converter
from ingest.importer.conversion.data_converter import DataType, CONVERTER_MAP, ListConverter
from ingest.template.exceptions import UnknownKeySchemaException
from ingest.template.schema_template import SchemaTemplate

_LOGGER = logging.getLogger(__name__)


class ConversionType(Enum):
    UNDEFINED = 0,
    MEMBER_FIELD = 1,
    FIELD_OF_LIST_ELEMENT = 2,
    IDENTITY = 3,
    LINKED_IDENTITY = 4,
    LINKED_EXTERNAL_REFERENCE = 5,
    LINKING_DETAIL = 6,
    EXTERNAL_REFERENCE = 7


class ColumnSpecification:
    UNKNOWN_DOMAIN_TYPE = '_unknown_type'

    # context_concrete_type is the concrete type of the Metadata Entity for which the column
    # represented by this object is specified for.
    def __init__(self, schema_template: SchemaTemplate, header: str, context_concrete_type, context=None,
                 order_of_occurrence=1):
        # Context refers to the context in which the header is being specified for.
        # For example, the property `project.contributors.email` will have a slightly different
        # specification in the context of `project.contributors`, than in the context of `project`.
        # In the former, email does not have a multivalue parent, whereas it has in the latter.
        # Framing it differently, in the former, it is each of the `contributors` that's being defined;
        # in the latter it is the `project` that's being defined.

        if not context:
            context = context_concrete_type

        parent_field, *_ = utils.split_field_chain(header)
        parent_spec = self._map_key_to_spec(schema_template, parent_field) if parent_field != context else None

        field_spec = self._map_key_to_spec(schema_template, header)
        data_type = DataType.find(field_spec.get('value_type'))

        # concrete_type is the actual concrete type that the header represents. Particularly in cases
        # where the column represents a linking detail to another type, concrete_type is different from
        # context_concrete_type. concrete_type is the "inherent" type of the column whichever context
        # it is specified in.
        concrete_type = utils.extract_root_field(header)
        type_spec = self._map_key_to_spec(schema_template, concrete_type)
        domain_entity_from_schema = None
        if type_spec:
            schema = type_spec.get('schema')
            domain_entity_from_schema, *_ = schema.get('domain_entity').split('/')

        self.field_name = header
        self.context_concrete_type = context_concrete_type
        self.domain_type = domain_entity_from_schema if domain_entity_from_schema else \
            ColumnSpecification.UNKNOWN_DOMAIN_TYPE
        self.data_type = data_type
        self.multivalue = field_spec.get('multivalue')
        self.multivalue_parent = (parent_spec and parent_spec.get('multivalue'))
        self.order_of_occurrence = order_of_occurrence
        self.entity_type = utils.extract_root_field(header)

        self.identity = field_spec.get('identifiable')
        self.external_reference = field_spec.get('external_reference')

    def is_multivalue(self):
        return self.multivalue

    def is_field_of_list_element(self):
        return self.multivalue_parent

    def is_identity(self):
        return self.identity

    def is_external_reference(self):
        return self.external_reference

    def get_conversion_type(self):
        if self._represents_an_object_field():
            conversion_type = self._determine_conversion_type_for_object_field()
        else:
            conversion_type = self._determine_conversion_type_for_reference_field()
        return conversion_type

    def _represents_an_object_field(self):
        return self.entity_type == self.context_concrete_type and self.order_of_occurrence == 1 and not \
            self.external_reference

    def _determine_conversion_type_for_object_field(self):
        if self.identity:
            conversion_type = ConversionType.IDENTITY
        elif self.multivalue_parent:
            conversion_type = ConversionType.FIELD_OF_LIST_ELEMENT
        else:
            conversion_type = ConversionType.MEMBER_FIELD
        return conversion_type

    def _determine_conversion_type_for_reference_field(self):
        if self.external_reference and self.entity_type == self.context_concrete_type:
            conversion_type = ConversionType.EXTERNAL_REFERENCE
        elif self.external_reference and self.entity_type != self.context_concrete_type:
            conversion_type = ConversionType.LINKED_EXTERNAL_REFERENCE
        elif self.identity:
            conversion_type = ConversionType.LINKED_IDENTITY
        else:
            conversion_type = ConversionType.LINKING_DETAIL
        return conversion_type

    def determine_converter(self):
        if not self.multivalue:
            converter = CONVERTER_MAP.get(self.data_type, data_converter.DEFAULT)
        else:
            converter = ListConverter(self.data_type)
        return converter

    @staticmethod
    def _map_key_to_spec(schema_template: SchemaTemplate, key):
        try:
            spec = schema_template.lookup_property_from_template(key)
        except UnknownKeySchemaException:
            _LOGGER.warning(f'[{key}] not found in the schema.')
            spec = {}
        return spec



import re
from abc import abstractmethod

from ingest.importer.conversion import data_converter
from ingest.importer.conversion.column_specification import \
    ColumnSpecification, ConversionType
from ingest.importer.conversion.data_converter import Converter, ListConverter
from ingest.importer.conversion.exceptions import UnknownMainCategory
from ingest.importer.conversion.metadata_entity import MetadataEntity
from ingest.importer.conversion.utils import split_field_chain

_LIST_CONVERTER = ListConverter()


class CellConversion(object):

    def __init__(self, field, converter: Converter):
        self.field = field
        self.applied_field = self._process_applied_field(field)
        self.converter = converter

    @staticmethod
    def _process_applied_field(field):
        pattern = r'(\w*\.){0,1}(?P<insert_field>.*)'
        match = re.match(pattern, field)
        return match.group('insert_field')

    @abstractmethod
    def apply(self, metadata: MetadataEntity, cell_data):
        # TODO Implement this method
        pass


class DirectCellConversion(CellConversion):

    def apply(self, metadata: MetadataEntity, cell_data):
        if cell_data is not None:
            content = self.converter.convert(cell_data)
            metadata.define_content(self.applied_field, content)


class ListElementCellConversion(CellConversion):

    def __init__(self, field: str, converter: Converter):
        list_converter = ListConverter(base_converter=converter)
        super(ListElementCellConversion, self).__init__(field, list_converter)

    def apply(self, metadata: MetadataEntity, cell_data):
        if cell_data is not None:
            parent_path, target_field = split_field_chain(self.applied_field)
            data_list = self.converter.convert(cell_data)
            parent = self._prepare_array(metadata, parent_path, len(data_list))
            for index, data in enumerate(data_list):
                target_object = parent[index]
                target_object[target_field] = data

    @staticmethod
    def _prepare_array(metadata, path, child_count):
        parent = metadata.get_content(path)
        if parent is None:
            parent = []
            metadata.define_content(path, parent)
        current_count = len(parent)
        missing_child_count = child_count - current_count
        if missing_child_count > 0:
            missing_children = [{} for _ in range(0, missing_child_count)]
            parent.extend(missing_children)
        return parent


class FieldOfSingleElementListCellConversion(CellConversion):

    def apply(self, metadata: MetadataEntity, cell_data):
        if cell_data is not None:
            parent_path, target_field = split_field_chain(self.applied_field)
            target_object = self._determine_target_object(metadata, parent_path)
            data = self.converter.convert(cell_data)
            target_object[target_field] = data

    @staticmethod
    def _determine_target_object(metadata, parent_path):
        parent = metadata.get_content(parent_path)
        if parent is None:
            target_object = {}
            parent = [target_object]
            metadata.define_content(parent_path, parent)
        else:
            target_object = parent[0]
        return target_object


class IdentityCellConversion(CellConversion):

    def apply(self, metadata: MetadataEntity, cell_data):
        value = self.converter.convert(cell_data)
        metadata.object_id = metadata.object_id or value
        metadata.define_content(self.applied_field, value)


class ExternalReferenceCellConversion(CellConversion):
    def apply(self, metadata: MetadataEntity, cell_data):
        value = self.converter.convert(cell_data)
        metadata._is_reference = True
        metadata.object_id = value


class LinkedIdentityCellConversion(CellConversion):

    def __init__(self, field, main_category):
        super(LinkedIdentityCellConversion, self).__init__(field, _LIST_CONVERTER)
        self.main_category = main_category

    def apply(self, metadata: MetadataEntity, cell_data):
        if self.main_category is None:
            raise UnknownMainCategory()
        if cell_data:
            links = self.converter.convert(cell_data)
            metadata.add_links(self.main_category, links)


class LinkedExternalReferenceCellConversion(CellConversion):

    def __init__(self, field, main_category):
        super(LinkedExternalReferenceCellConversion, self).__init__(field, _LIST_CONVERTER)
        self.main_category = main_category

    def apply(self, metadata: MetadataEntity, cell_data):
        if cell_data:
            links = self.converter.convert(cell_data)
            metadata.add_external_links(self.main_category, links)


class LinkingDetailCellConversion(CellConversion):

    def apply(self, metadata: MetadataEntity, cell_data):
        value = self.converter.convert(cell_data)
        metadata.define_linking_detail(self.applied_field, value)


class DoNothing(CellConversion):

    def __init__(self):
        super(DoNothing, self).__init__('', data_converter.DEFAULT)

    def apply(self, metadata: MetadataEntity, cell_data):
        pass


DO_NOTHING = DoNothing()


def determine_strategy(column_spec: ColumnSpecification):
    strategy = DO_NOTHING
    if column_spec is not None:
        field_name = column_spec.field_name
        converter = column_spec.determine_converter()
        conversion_type = column_spec.get_conversion_type()
        if ConversionType.MEMBER_FIELD == conversion_type:
            strategy = DirectCellConversion(field_name, converter)
        elif ConversionType.FIELD_OF_LIST_ELEMENT == conversion_type:
            strategy = ListElementCellConversion(field_name, converter)
        elif ConversionType.LINKING_DETAIL == conversion_type:
            strategy = LinkingDetailCellConversion(field_name, converter)
        elif ConversionType.IDENTITY == conversion_type:
            strategy = IdentityCellConversion(field_name, converter)
        elif ConversionType.LINKED_IDENTITY == conversion_type:
            strategy = LinkedIdentityCellConversion(field_name, column_spec.domain_type)
        elif ConversionType.LINKED_EXTERNAL_REFERENCE == conversion_type:
            strategy = LinkedExternalReferenceCellConversion(field_name, column_spec.domain_type)
        elif ConversionType.EXTERNAL_REFERENCE == conversion_type:
            strategy = ExternalReferenceCellConversion(field_name, converter)
    return strategy



import os

from setuptools import setup, find_packages

base_dir = os.path.dirname(__file__)
install_requires = [line.rstrip() for line in open(os.path.join(base_dir, 'requirements.txt'))]

setup(
    name='hca_ingest',
    version='0.6.10',
    packages=find_packages(exclude=['tests', 'tests.*']),
    install_requires=install_requires,
    include_package_data=True
)



from requests import Session
from requests.adapters import HTTPAdapter
from urllib3.util import retry


def create_session_with_retry(retry_policy=None) -> Session:
    retry_policy = retry_policy or retry.Retry(
        total=50,
        # seems that this has a default value of 10,
        # setting this to a very high number so that it'll respect the status retry count

        status=10,
        # status is the no. of retries if response is in status_forcelist

        read=10,
        status_forcelist=[409],
        backoff_factor=0.6,
        method_whitelist=frozenset(
            ['HEAD', 'GET', 'POST', 'PUT', 'DELETE', 'OPTIONS', 'TRACE'])
    )
    session = Session()
    adapter = HTTPAdapter(max_retries=retry_policy)
    session.mount('http://', adapter)
    session.mount('https://', adapter)
    return session



import re

SPLIT_FIELD_PATTERN = re.compile(r'^(?P<parent>\w*(\.\w*)*)\.(?P<target>\w*)$')


def split_field_chain(field):
    match = SPLIT_FIELD_PATTERN.search(field)

    parent_path = match.group('parent') if match else ''
    target_field = match.group('target') if match else field

    return parent_path, target_field


def extract_root_field(field_chain):
    root_field = None
    if field_chain is not None:
        split = field_chain.split('.')
        root_field = split[0]
    return root_field



#!/usr/bin/env python
"""
Description goes here
"""
import datetime
import json
import logging
import os
import time

import hca
from hca.util import SwaggerAPIException

from ingest.api import utils
from ingest.utils.token_manager import Token

__author__ = "jupp"
__license__ = "Apache 2.0"
__date__ = "12/09/2017"

AUTH_INFO_ENV_VAR = "EXPORTER_AUTH_INFO"


class DssApi:
    def __init__(self, url=None):
        format = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
        logging.basicConfig(format=format)
        logging.getLogger("requests").setLevel(logging.WARNING)
        logging.getLogger("requests").setLevel(logging.INFO)
        self.logger = logging.getLogger(__name__)

        self.url = url if url else "https://dss.dev.data.humancellatlas.org"
        if not url and 'DSS_API' in os.environ:
            url = os.environ['DSS_API']
            # expand interpolated env vars
            self.url = os.path.expandvars(url)
            self.logger.info("using " + url + " for dss API")

        self.headers = {'Content-type': 'application/json'}

        self.dss_client = None

        self.token = None
        self.init_dss_client()

    # TODO This is workaround to not let the dss client token expired
    # See ticket https://app.zenhub.com/workspaces/dcp-5ac7bcf9465cb172b77760d9/issues/humancellatlas/data-store/2216
    # Create a dummy token at the same time DSS client is initialised.
    # This is just a way to keep track when DSS client token will be expired
    def init_dss_client(self, duration=3600 * 1000, refresh_period=60 * 10 * 1000):
        if not self.token or self.token.is_expired():
            self.token = Token('dummy', duration, refresh_period)
            self.dss_client = hca.dss.DSSClient(
                swagger_url=f'{self.url}/v1/swagger.json')
            self.dss_client.host = self.url + "/v1"
            self.creator_uid = 8008

    def put_file(self, bundle_uuid, file):
        url = file["url"]
        uuid = file["dss_uuid"]

        update_date = file.get("update_date")
        version = utils.to_dss_version(update_date) if update_date \
            else datetime.datetime.utcnow().strftime("%Y-%m-%dT%H%M%S.%fZ")

        # retrying file creation 20 times
        max_retries = 20
        tries = 0
        file_create_complete = False

        params = {
            'uuid': uuid,
            'version': version,
            'creator_uid': self.creator_uid,
            'source_url': url
        }

        if bundle_uuid:
            params["bundle_uuid"] = bundle_uuid

        while not file_create_complete and tries < max_retries:
            try:
                tries += 1
                self.logger.info(f'Creating file in DSS {uuid}:{version} with params: {json.dumps(params)}')
                bundle_file = None

                if bundle_uuid:
                    self.init_dss_client()
                    bundle_file = self.dss_client.put_file(
                        uuid=uuid,
                        version=version,
                        bundle_uuid=bundle_uuid,
                        creator_uid=self.creator_uid,
                        source_url=url
                    )
                else:
                    self.init_dss_client()
                    bundle_file = self.dss_client.put_file(
                        uuid=uuid,
                        version=version,
                        creator_uid=self.creator_uid,
                        source_url=url
                    )

                self.logger.info('Created!')
                file_create_complete = True
                return bundle_file
            except Exception as e:
                self.logger.error(
                    'Attempt {0} out of {1}: Error in hca_client.put_file method call with params:{2} due to {'
                    '3}'.format(
                        str(tries),
                        str(max_retries),
                        json.dumps(params),
                        str(e))
                )

                if not tries < max_retries:
                    raise Error(e)
                else:
                    time.sleep(60)

    def put_bundle(self, bundle_uuid, version, bundle_files):
        bundle = None

        # retrying file creation 20 times
        max_retries = 20
        tries = 0
        bundle_create_complete = False

        # finally create the bundle
        while not bundle_create_complete and tries < max_retries:
            try:
                tries += 1
                self.logger.info(f'Creating bundle in DSS {bundle_uuid}:{version}')
                self.init_dss_client()
                bundle = self.dss_client.put_bundle(
                    uuid=bundle_uuid,
                    version=version,
                    replica="aws",
                    files=bundle_files,
                    creator_uid=self.creator_uid
                )
                self.logger.info('Created!')
                bundle_create_complete = True
                return bundle

            except Exception as e:
                if isinstance(e, SwaggerAPIException):
                    error_code = e.details.get('code')

                    if error_code.lower() == 'bundle_already_exists':
                        raise BundleAlreadyExist(f"Bundle {bundle_uuid}.{version} already exist in DSS.")

                params = {
                    'uuid': bundle_uuid,
                    'version': version,
                    'replica': "aws",
                    'files': bundle_files,
                    'creator_uid': self.creator_uid
                }
                self.logger.error(
                    'Attempt {0} out of {1}: Error in hca_client.put_bundle method call with params:{2} due to {'
                    '3}'.format(
                        str(tries),
                        str(max_retries),
                        json.dumps(params),
                        str(e))
                )
                if not tries < max_retries:
                    raise Error(e)
                else:
                    time.sleep(60)

    def get_bundle(self, bundle_uuid, version=None):
        self.init_dss_client()
        if version:
            bundle = self.dss_client.get_bundle(
                uuid=bundle_uuid,
                replica="aws",
                version=version
            )
            bundle_files = self.get_bundle_files(bundle_uuid, version=version)
        else:
            bundle = self.dss_client.get_bundle(
                uuid=bundle_uuid,
                replica="aws",
                version=version
            )
            bundle_files = self.get_bundle_files(bundle_uuid, version=version)
        bundle["bundle"]["files"] = bundle_files
        return bundle

    def get_bundle_files(self, bundle_uuid, version=None):
        if version:
            return list(self.dss_client.get_bundle.iterate(
                uuid=bundle_uuid,
                replica="aws",
                version=version
            ))
        else:
            return list(self.dss_client.get_bundle.iterate(
                uuid=bundle_uuid,
                replica="aws"
            ))

    def get_file(self, file_uuid, version=None):
        self.init_dss_client()
        if version:
            return self.dss_client.get_file(
                uuid=file_uuid,
                replica="aws",
                version=version
            )
        else:
            return self.dss_client.get_file(
                uuid=file_uuid,
                replica="aws"
            )

    def head_file(self, file_uuid, version=None):
        self.init_dss_client()
        # finally create the bundle
        try:
            return self.dss_client.head_file(
                uuid=file_uuid,
                version=version,
                replica="aws"
            )
        except Exception as e:
            raise Error(e)


# Module Exceptions

class Error(Exception):
    """Base-class for all exceptions raised by this module."""


class BundleAlreadyExist(Error):
    """Bundle in DSS already Exist."""



import base64
import functools
import json
import time
from datetime import datetime

import jwt
import requests
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric import rsa


class DCPAuthClient(object):
    audience = "https://dev.data.humancellatlas.org/"
    dev_deployments = ('dev', 'integration', 'test', 'staging')

    def __init__(self, trusted_google_project, jwt_info):
        self.auth_jwt_info = jwt_info
        self.trusted_google_project = trusted_google_project

        if not any(deployment in trusted_google_project for deployment in DCPAuthClient.dev_deployments):
            self.audience = "https://data.humancellatlas.org/"

        self.__token = None
        # TODO: check the liveness of the token and make it as singleton if applicable
        self.issue_time = None
        self.expire_time = None

    def get_auth_header(self):
        return {'Authorization': 'Bearer {}'.format(self.token)}

    @property
    def token(self):
        credentials = self.auth_jwt_info
        tok = DCPAuthClient.get_service_jwt(service_credentials=credentials, audience=self.audience)
        self.verify_jwt(tok, audience=self.audience, trusted_google_project=self.trusted_google_project)
        return self.__token

    @staticmethod
    def _from_json(file_path):
        with open(file_path, 'r') as f:
            return json.load(f)

    @staticmethod
    @functools.lru_cache(maxsize=32)
    def get_openid_config(openid_provider):
        res = requests.get("https://{}/.well-known/openid-configuration".format(openid_provider))
        res.raise_for_status()
        return res.json()

    @staticmethod
    def get_jwks_uri(openid_provider):
        if openid_provider.endswith("iam.gserviceaccount.com"):
            return "https://www.googleapis.com/service_accounts/v1/jwk/{}".format(openid_provider)
        else:
            return DCPAuthClient.get_openid_config(openid_provider)["jwks_uri"]

    @staticmethod
    @functools.lru_cache(maxsize=32)
    def get_public_keys(openid_provider):
        keys = requests.get(DCPAuthClient.get_jwks_uri(openid_provider)).json()["keys"]
        return {
            key["kid"]: rsa.RSAPublicNumbers(
                e=int.from_bytes(base64.urlsafe_b64decode(key["e"] + "==="), byteorder="big"),
                n=int.from_bytes(base64.urlsafe_b64decode(key["n"] + "==="), byteorder="big")
            ).public_key(backend=default_backend())
            for key in keys
        }

    @staticmethod
    def get_service_jwt(service_credentials, audience):
        iat = time.time()
        exp = iat + 3600
        payload = {'iss': service_credentials["client_email"],
                   'sub': service_credentials["client_email"],
                   'aud': audience,
                   'iat': iat,
                   'exp': exp,
                   'https://auth.data.humancellatlas.org/email': service_credentials["client_email"],
                   'https://auth.data.humancellatlas.org/group': 'hca',
                   'scope': ["openid", "email", "offline_access"]
                   }
        additional_headers = {'kid': service_credentials["private_key_id"]}
        signed_jwt = jwt.encode(payload, service_credentials["private_key"], headers=additional_headers,
                                algorithm='RS256').decode()
        return signed_jwt

    def verify_jwt(self, token, audience, trusted_google_project):
        assert token
        try:
            openid_provider = "humancellatlas.auth0.com"
            token_header = jwt.get_unverified_header(token)
            public_keys = DCPAuthClient.get_public_keys(openid_provider)
            tok = jwt.decode(token, key=public_keys[token_header["kid"]], audience=audience)
        except KeyError:
            unverified_token = jwt.decode(token, verify=False)
            issuer = unverified_token["iss"]
            assert issuer.endswith("@{}.iam.gserviceaccount.com".format(trusted_google_project))
            token_header = jwt.get_unverified_header(token)
            public_keys = DCPAuthClient.get_public_keys(issuer)
            tok = jwt.decode(token, key=public_keys[token_header["kid"]], audience=audience)
        self.__token = token
        self.issue_time = datetime.fromtimestamp(tok['iat']).strftime('%Y-%m-%d %H:%M:%S')
        self.expire_time = datetime.fromtimestamp(tok['exp']).strftime('%Y-%m-%d %H:%M:%S')
